
from libcpp.string cimport string
from libcpp.vector cimport vector
from libc.stdint cimport uint32_t, int64_t
from libc.math cimport sqrt
from libcpp cimport bool

from ttcrpy.common cimport sxz, sxyz, siv, siv2, sijv, Node3Dc, Node3Dcsp, \
Node3Dn, Node3Dnsp, Cell, Node2Dcsp, Node2Dn, Node2Dnsp


cdef extern from "typedefs.h" namespace "ttcr":
    cdef cppclass node2d:
        pass
    cdef cppclass cell2d:
        pass
    cdef cppclass cell2d_e:
        pass
    cdef cppclass cell2d_te:
        pass
    cdef cppclass cell2d_p:
        pass
    cdef cppclass cell2d_h:
        pass


cdef extern from "Grid3D.h" namespace "ttcr" nogil:
    cdef cppclass Grid3D[T1,T2]:
        void setUsePool(bool up)
        size_t getNthreads()
        void setSlowness(vector[T1]&) except +
        void getSlowness(vector[T1]&) except +
        T1 computeSlowness(sxyz[T1]&) except +
        void getTT(vector[T1]& tt, size_t threadNo) except +
        void raytrace(vector[sxyz[T1]]& Tx,
                      vector[T1]& t0,
                      vector[sxyz[T1]]& Rx,
                      vector[T1]& tt,
                      size_t thread_no) except +
        void raytrace(vector[sxyz[T1]]& Tx,
                      vector[T1]& t0,
                      vector[sxyz[T1]]& Rx,
                      vector[T1]& tt,
                      vector[vector[sxyz[T1]]]& r_data,
                      size_t thread_no) except +
        void raytrace(vector[sxyz[T1]]& Tx,
                      vector[T1]& t0,
                      vector[sxyz[T1]]& Rx,
                      vector[T1]& tt,
                      vector[vector[siv[T1]]]& l_data,
                      size_t thread_no) except +
        void raytrace(vector[sxyz[T1]]& Tx,
                      vector[T1]& t0,
                      vector[sxyz[T1]]& Rx,
                      vector[T1]& tt,
                      vector[vector[sxyz[T1]]]& r_data,
                      vector[vector[siv[T1]]]& l_data,
                      size_t thread_no) except +
        void raytrace(vector[sxyz[T1]]& Tx,
                      vector[T1]& t0,
                      vector[sxyz[T1]]& Rx,
                      vector[T1]& tt,
                      vector[vector[sijv[T1]]]& m_data,
                      size_t thread_no) except +
        void raytrace(vector[sxyz[T1]]& Tx,
                      vector[T1]& t0,
                      vector[sxyz[T1]]& Rx,
                      vector[T1]& tt,
                      vector[vector[sxyz[T1]]]& r_data,
                      vector[vector[sijv[T1]]]& m_data,
                      size_t thread_no) except +
        void raytrace(vector[vector[sxyz[T1]]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[sxyz[T1]]]& Rx,
                      vector[vector[T1]]& traveltimes) except +
        void raytrace(vector[vector[sxyz[T1]]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[sxyz[T1]]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[sxyz[T1]]]]& r_data) except +
        void raytrace(vector[vector[sxyz[T1]]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[sxyz[T1]]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[sijv[T1]]]]& m_data) except +
        void raytrace(vector[vector[sxyz[T1]]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[sxyz[T1]]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[sxyz[T1]]]]& r_data,
                      vector[vector[vector[sijv[T1]]]]& m_data) except +
        void raytrace(vector[vector[sxyz[T1]]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[sxyz[T1]]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[siv[T1]]]]& l_data) except +
        void raytrace(vector[vector[sxyz[T1]]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[sxyz[T1]]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[sxyz[T1]]]]& r_data,
                      vector[vector[vector[siv[T1]]]]& l_data) except +


cdef extern from "Grid3Drn.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Drn[T1,T2,N](Grid3D[T1,T2]):
        pass

cdef extern from "Grid3Drc.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Drc[T1,T2,N](Grid3D[T1,T2]):
        pass

cdef extern from "Grid3Drnfs.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Drnfs[T1,T2](Grid3Drn[T1,T2,Node3Dn[T1,T2]]):
        Grid3Drnfs(T2, T2, T2, T1, T1, T1, T1, T1, int, bool, bool, bool,
                   size_t) except +

cdef extern from "Grid3Drnsp.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Drnsp[T1,T2](Grid3Drn[T1,T2,Node3Dnsp[T1,T2]]):
        Grid3Drnsp(T2, T2, T2, T1, T1, T1, T1, T1, T1, T2, T2, T2, bool, bool,
                   size_t) except +

cdef extern from "Grid3Drndsp.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Drndsp[T1,T2](Grid3Drn[T1,T2,Node3Dn[T1,T2]]):
        Grid3Drndsp(T2, T2, T2, T1, T1, T1, T1, T1, T1, T2, bool, T2, T1, bool,
                    size_t) except +

cdef extern from "Grid3Drcfs.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Drcfs[T1,T2](Grid3Drn[T1,T2,Node3Dn[T1,T2]]):
        Grid3Drcfs(T2, T2, T2, T1, T1, T1, T1, T1, int, bool, bool, bool,
                   size_t) except +

cdef extern from "Grid3Drcsp.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Drcsp[T1,T2,CELL](Grid3Drc[T1,T2,Node3Dcsp[T1,T2]]):
        Grid3Drcsp(T2, T2, T2, T1, T1, T1, T1, T1, T1, T2, T2, T2, bool,
                   size_t) except +

cdef extern from "Grid3Drcdsp.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Drcdsp[T1,T2,CELL](Grid3Drc[T1,T2,Node3Dc[T1,T2]]):
        Grid3Drcdsp(T2, T2, T2, T1, T1, T1, T1, T1, T1, T2, bool, T2, T1,
                    size_t) except +


cdef extern from "Grid2D.h" namespace "ttcr" nogil:
    cdef cppclass Grid2D[T1,T2,S]:
        size_t getNthreads()
        void setSlowness(vector[T1]&) except +
        void getSlowness(vector[T1]&) except +
        void setXi(vector[T1]&) except +
        void setTiltAngle(vector[T1]&) except +
        void setVp0(vector[T1]&) except +
        void setVs0(vector[T1]&) except +
        void setDelta(vector[T1]&) except +
        void setEpsilon(vector[T1]&) except +
        void setGamma(vector[T1]&) except +
        void getTT(vector[T1]& tt, size_t threadNo) except +
        void raytrace(vector[S]& Tx,
                      vector[T1]& t0,
                      vector[S]& Rx,
                      vector[T1]& traveltimes,
                      size_t threadNo) except +
        void raytrace(vector[S]& Tx,
                      vector[T1]& t0,
                      vector[S]& Rx,
                      vector[T1]& traveltimes,
                      vector[vector[sxz[T1]]]& r_data,
                      size_t threadNo) except +
        void raytrace(vector[S]& Tx,
                      vector[T1]& t0,
                      vector[S]& Rx,
                      vector[T1]& traveltimes,
                      vector[vector[siv2[T1]]]& l_data,
                      size_t threadNo) except +
        void raytrace(vector[S]& Tx,
                      vector[T1]& t0,
                      vector[S]& Rx,
                      vector[T1]& traveltimes,
                      vector[vector[S]]& r_data,
                      vector[vector[siv2[T1]]]& l_data,
                      size_t threadNo) except +
        void raytrace(vector[vector[S]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[S]]& Rx,
                      vector[vector[T1]]& traveltimes) except +
        void raytrace(vector[vector[S]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[S]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[S]]]& r_data) except +
        void raytrace(vector[vector[S]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[S]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[siv2[T1]]]]& l_data) except +
        void raytrace(vector[vector[S]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[S]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[S]]]& r_data,
                      vector[vector[vector[siv2[T1]]]]& l_data) except +

cdef extern from "Grid2Drc.h" namespace "ttcr" nogil:
    cdef cppclass Grid2Drc[T1,T2,S,NODE,CELL](Grid2D[T1,T2,S]):
        pass

cdef extern from "Grid2Drn.h" namespace "ttcr" nogil:
    cdef cppclass Grid2Drn[T1,T2,S,NODE](Grid2D[T1,T2,S]):
        pass

cdef extern from "Grid2Drcsp.h" namespace "ttcr" nogil:
    cdef cppclass Grid2Drcsp[T1,T2,S,CELL](Grid2Drc[T1,T2,S,node2d,CELL]):
        Grid2Drcsp(T2, T2, T1, T1, T1, T1, T2, T2, size_t) except +

cdef extern from "Grid2Drcfs.h" namespace "ttcr" nogil:
    cdef cppclass Grid2Drcfs[T1,T2,S](Grid2Drn[T1,T2,S,Node2Dn[T1,T2]]):
        Grid2Drcfs(T2, T2, T1, T1, T1, T1, T1, int, bool, bool, size_t) except +

cdef extern from "Grid2Drnsp.h" namespace "ttcr" nogil:
    cdef cppclass Grid2Drnsp[T1,T2,S](Grid2Drn[T1,T2,S,node2d]):
        Grid2Drnsp(T2, T2, T1, T1, T1, T1, T2, T2, size_t) except +

cdef extern from "Grid2Drnfs.h" namespace "ttcr" nogil:
    cdef cppclass Grid2Drnfs[T1,T2,S](Grid2Drn[T1,T2,S,Node2Dn[T1,T2]]):
        Grid2Drnfs(T2, T2, T1, T1, T1, T1, T1, int, bool, bool, size_t) except +



# -*- coding: utf-8 -*-
"""
.. deprecated:: v0.4.0
    `cgrid2d` will be removed in ttcrpy v1.0.0, it is replaced by
    `rgrid`.

    2D grid for raytracing based on the shortest path method

    Notes
    -----
        The raytracing codes are based on a column-major order
        for the slowness vector (Z is the "fast" axis).
        To visualize the slowness model with Z axis vertical and X horizontal,
        the vector should be reshaped as
        slowness.reshape(nx,nz).T

    This code is part of ttcr ( https://github.com/groupeLIAMG/ttcr )
"""

"""
    Copyright 2016 Bernard Giroux
    email: bernard.giroux@ete.inrs.ca

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


from libcpp.string cimport string
from libcpp.vector cimport vector
from libc.stdint cimport uint32_t

import numpy as np
cimport numpy as np

from scipy.sparse import csr_matrix

cdef extern from "ttcr_t.h" namespace "ttcr":
    cdef cppclass sxz[T]:
        sxz(T, T) except +


cdef extern from "Grid2Dttcr.h" namespace "ttcr":
    cdef cppclass Grid2Dttcr:
        Grid2Dttcr(string&, uint32_t, uint32_t, double, double, double, double, uint32_t, uint32_t, size_t) except +
        string getType()
        void setSlowness(const vector[double]&) except +
        void setXi(const vector[double]&) except +
        void setTheta(const vector[double]&) except +
        void raytrace(vector[sxz[double]]&,vector[double]&,vector[sxz[double]]&,double*,object,object) except +
        void raytrace(vector[sxz[double]]&,vector[double]&,vector[sxz[double]]&,double*,object) except +
        void raytrace(vector[sxz[double]]&,vector[double]&,vector[sxz[double]]&,double*) except +
        @staticmethod
        int Lsr2d(double*,double*,size_t,double*,size_t,double*,size_t,object)
        @staticmethod
        int Lsr2da(double*,double*,size_t,double*,size_t,double*,size_t,object)


cdef extern from "Cell.h" namespace "ttcr":
    cdef cppclass Cell[T, NODE, S]:
        Cell(size_t)
    cdef cppclass CellElliptical[T, NODE, S]:
        CellElliptical(size_t)
    cdef cppclass CellTiltedElliptical[T, NODE, S]:
        CellTiltedElliptical(size_t)


cdef extern from "Node2Dcsp.h" namespace "ttcr":
    cdef cppclass Node2Dcsp[T1, T2]:
        Node2Dcsp(size_t)


cdef extern from "Grid2Drcsp.h" namespace "ttcr":
    cdef cppclass Grid2Drcsp[T1, T2, S, CELL]:
        Grid2Drcsp(T2, T2, T1, T1, T1, T1, T2, T2, size_t)
        size_t getNthreads()
        void setSlowness(vector[T1]&) except +


cdef class Grid2Dcpp:
    """
    Grid2Dcpp(type, nx, nz, dx, dz, xmin, zmin, nsnx, nsnz, nthreads)

    Parameters
    ----------
    type : type of media
             'iso'          isotorpic
             'elliptical'   elliptically anisotropic
             'tilted'       tilted elliptically anisotropic
    nx : number of cells in x
    nz : number of cells in z
    dx : cell size in x
    dz : cell size in z
    xmin : origin inx
    zmin : origin in z
    nsnx : number of secondary nodes in x
    nsnz : number of secondary nodes in z
    nthreads : number of threads
    """
    cdef Grid2Dttcr* grid
    def __cinit__(self, gridType, uint32_t nx, uint32_t nz, double dx, double dz,
                  double xmin, double zmin, uint32_t nsnx, uint32_t nsnz,
                  size_t nthreads):
        self.grid = new Grid2Dttcr(gridType, nx, nz, dx, dz, xmin, zmin, nsnx, nsnz, nthreads)

    def __dealloc__(self):
        del self.grid

    def getType(self):
        return self.grid.getType()

    def raytrace(self, slowness, xi, theta, Tx, Rx, t0, nout):
        """
        raytrace(slowness, xi, theta, Tx, Rx, t0, nout) -> tt,L,rays

        Parameters
        ----------
        slowness : vector of slowness at grid cells
        xi : vector of anisotropy ratio at grid cells
        theta : vector of anisotropy angle at grid cells
        Tx : source coordinates, nTx by 2
               1st column contains X coordinates,
               2nd contains Z coordinates
        Rx : receiver coordinates, nTx by 2
               1st column contains X coordinates,
               2nd contains Z coordinates
        t0 : origin time
        nout : number of output variables (1, 2 or 3)

        Returns
        -------
        tt : travel times
        L : (if nout > 1) data kernel matrix (tt = L*slowness)
        rays : (if nout == 3) coordinates of rays
        """

        # check if types are consistent with input data
        if len(xi) != 0:
            if len(theta) != 0:
                if self.grid.getType() != b'tilted':
                    raise TypeError('Grid should handle raytracing in tilted elliptically anisotropic media')
            else:
                if self.grid.getType() != b'elliptical':
                    raise TypeError('Grid should handle raytracing in elliptically anisotropic media')
        else:
            if self.grid.getType() != b'iso':
                raise TypeError('Grid should handle raytracing in isotropic media')

        # assing model data
        cdef vector[double] slown# = slowness.tolist()
        for tmp in slowness:
            slown.push_back(tmp)
        self.grid.setSlowness(slown)
        cdef vector[double] x
        if len(xi) != 0:
            for tmp in xi:
                x.push_back(tmp)
            self.grid.setXi(x)
        cdef vector[double] t
        if len(theta) != 0:
            for tmp in theta:
                t.push_back(tmp)
            self.grid.setTheta(t)

        # create C++ input variables
        cdef vector[sxz[double]] cTx
        for t in Tx:
            cTx.push_back(sxz[double](t[0], t[1]))

        cdef vector[sxz[double]] cRx
        for r in Rx:
            cRx.push_back(sxz[double](r[0], r[1]))

        cdef vector[double] ct0# = t0.tolist()
        for tmp in t0:
            ct0.push_back(tmp)

        # instantiate output variables
        cdef np.ndarray tt = np.empty([Rx.shape[0],], dtype=np.double)  # tt should be the right size

        if nout==1:
            self.grid.raytrace(cTx, ct0, cRx, <double*> np.PyArray_DATA(tt))
            return tt

        elif nout==2:
            Ldata = ([0.0], [0.0], [0.0])
            self.grid.raytrace(cTx, ct0, cRx, <double*> np.PyArray_DATA(tt), Ldata)

            M = Rx.shape[0]
            N = len(slowness)
            if self.grid.getType() != b'iso':
                N = 2*N

            L = csr_matrix(Ldata, shape=(M,N))

            return tt,L

        elif nout==3:
            rays = tuple([ [0.0] for i in range(Rx.shape[0]) ])
            Ldata = ([0.0], [0.0], [0.0])

            self.grid.raytrace(cTx, ct0, cRx, <double*> np.PyArray_DATA(tt), rays, Ldata)

            M = Rx.shape[0]
            N = len(slowness)
            if self.grid.getType() != b'iso':
                N = 2*N

            L = csr_matrix(Ldata, shape=(M,N))

            return tt,L,rays


    @staticmethod
    def Lsr2d(Tx, Rx, grx, grz):
        """
        Lsr2d(Tx, Rx, grx, grz) -> L

        Raytracing with straight rays in 2D

        Parameters
        ----------
        Tx : source coordinates, nTx by 2
              1st column contains X coordinates,
              2nd contains Z coordinates
        Rx : receiver coordinates, nTx by 2
              1st column contains X coordinates,
              2nd contains Z coordinates
        grx : grid node coordinates along x
        grz : grid node coordinates along z

        Returns
        -------
        L : data kernel matrix (tt = L*slowness)
        """

        cdef size_t nTx = Tx.shape[0]
        cdef size_t n_grx = grx.shape[0]
        cdef size_t n_grz = grz.shape[0]

        Ldata = ([0.0], [0.0], [0.0])

        Grid2Dttcr.Lsr2d(<double*> np.PyArray_DATA(Tx), <double*> np.PyArray_DATA(Rx), nTx, <double*> np.PyArray_DATA(grx), n_grx, <double*> np.PyArray_DATA(grz), n_grz, Ldata)

        M = nTx
        N = (n_grx-1)*(n_grz-1)
        L = csr_matrix(Ldata, shape=(M,N))

        return L


    @staticmethod
    def Lsr2da(Tx, Rx, grx, grz):
        """
        Lsr2da(Tx, Rx, grx, grz) -> L

        Raytracing with straight rays in 2D elliptically anisotropic media

        Parameters
        ----------
        Tx : source coordinates, nTx by 2
               1st column contains X coordinates,
               2nd contains Z coordinates
        Rx : receiver coordinates, nTx by 2
               1st column contains X coordinates,
               2nd contains Z coordinates
        grx : grid node coordinates along x
        grz : grid node coordinates along z

        Returns
        -------
        L : data kernel matrix (tt = L*slowness)
        """

        cdef size_t nTx = Tx.shape[0]
        cdef size_t n_grx = grx.shape[0]
        cdef size_t n_grz = grz.shape[0]

        Ldata = ([0.0], [0.0], [0.0])

        Grid2Dttcr.Lsr2da(<double*> np.PyArray_DATA(Tx), <double*> np.PyArray_DATA(Rx), nTx, <double*> np.PyArray_DATA(grx), n_grx, <double*> np.PyArray_DATA(grz), n_grz, Ldata)

        M = nTx
        N = 2*(n_grx-1)*(n_grz-1)
        L = csr_matrix(Ldata, shape=(M,N))

        return L


cdef class Grid2Diso:
    cdef uint32_t nx
    cdef uint32_t nz
    cdef Grid2Drcsp[double, uint32_t, sxz[double], Cell[double, Node2Dcsp[double, uint32_t], sxz[double]]]* grid

    def __cinit__(self, uint32_t nx, uint32_t nz, double dx, double dz,
                  double xmin, double zmin, uint32_t nsnx, uint32_t nsnz,
                  size_t nthreads):
        self.grid = new Grid2Drcsp[double, uint32_t, sxz[double], Cell[double, Node2Dcsp[double, uint32_t], sxz[double]]](nx, nz, dx, dz, xmin, zmin, nsnx, nsnz, nthreads)

    def __dealloc__(self):
        del self.grid

    def get_type(self):
        return 'iso'



# -*- coding: utf-8 -*-
"""
.. deprecated:: v0.4.0
    `cgrid3d` will be removed in ttcrpy v1.0.0, it is replaced by
    `rgrid`.

    3D grid for raytracing based on the fast sweeping method

    This code is part of ttcr ( https://github.com/groupeLIAMG/ttcr )
"""

"""
Copyright 2017 Bernard Giroux
email: bernard.giroux@ete.inrs.ca

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from libcpp.string cimport string
from libcpp.vector cimport vector
from libc.stdint cimport uint32_t
from libcpp cimport bool

import numpy as np
cimport numpy as np

from scipy.sparse import csr_matrix

cdef extern from "ttcr_t.h" namespace "ttcr":
    cdef cppclass sxyz[T]:
        sxyz(T, T, T) except +
        T x
        T y
        T z

    cdef cppclass siv[T]:
        siv(size_t, T) except +
        size_t i
        T v

    cdef cppclass sijv[T]:
        sijv(size_t, T) except +
        size_t i
        size_t j
        T v

cdef extern from "Grid3Drnfs.h" namespace "ttcr":
    cdef cppclass Grid3Drnfs[T1,T2]:
        Grid3Drnfs(T2, T2, T2, T1, T1, T1, T1, T1, int, bool, bool, bool, size_t) except +
        size_t getNthreads()
        void setSlowness(vector[T1]&) except +
        void raytrace(vector[sxyz[T1]]&,
                      vector[T1]&,
                      vector[sxyz[T1]]&,
                      vector[T1]&,
                      size_t) except +
        void raytrace(vector[sxyz[T1]]&,
                      vector[T1]&,
                      vector[sxyz[T1]]&,
                      vector[T1]&,
                      vector[vector[sxyz[T1]]]&,
                      size_t) except +
        void raytrace(vector[sxyz[T1]]&,
                      vector[T1]&,
                      vector[sxyz[T1]]&,
                      vector[T1]&,
                      vector[vector[sxyz[T1]]]&,
                      vector[vector[sijv[T1]]]&,
                      size_t) except +
        T1 computeSlowness(sxyz[T1]&)

cdef extern from "Grid3Drcfs.h" namespace "ttcr":
    cdef cppclass Grid3Drcfs[T1,T2]:
        Grid3Drcfs(T2, T2, T2, T1, T1, T1, T1, T1, int, bool, bool, bool, size_t) except +
        size_t getNthreads()
        void setSlowness(vector[T1]&) except +
        void raytrace(vector[sxyz[T1]]&,
                      vector[T1]&,
                      vector[sxyz[T1]]&,
                      vector[T1]&,
                      size_t) except +
        void raytrace2(vector[sxyz[T1]]&,
                       vector[T1]&,
                       vector[sxyz[T1]]&,
                       vector[T1]&,
                       vector[vector[sxyz[T1]]]&,
                       size_t) except +
        void raytrace(vector[sxyz[T1]]&,
                      vector[T1]&,
                      vector[sxyz[T1]]&,
                      vector[T1]&,
                      vector[vector[siv[T1]]]&,
                      size_t) except +
        void raytrace(vector[sxyz[T1]]&,
                      vector[T1]&,
                      vector[sxyz[T1]]&,
                      vector[T1]&,
                      vector[vector[sxyz[T1]]]&,
                      vector[vector[siv[T1]]]&,
                      size_t) except +

cdef class Grid3Drn:
    """
    Grid3Drn(nx, ny, nz, dx, xmin, ymin, zmin, eps, maxit, weno, nthreads)

    3D rectilinear grid with slowness defined at nodes

    Parameters
    ----------
    nx : number of cells in x
    ny : number of cells in y
    nz : number of cells in z
    dx : cell size in x
    xmin : origin in x
    ymin : origin in y
    zmin : origin in z
    eps : convergence criterion (FSM)
    maxit : max number of sweeping iterations
    weno : use 3rd order weighted essentially non-oscillatory operator (bool)
    nthreads : number of threads for raytracing
    """
    cdef uint32_t nx
    cdef uint32_t ny
    cdef uint32_t nz
    cdef Grid3Drnfs[double, uint32_t]* grid
    def __cinit__(self, uint32_t nx, uint32_t ny, uint32_t nz, double dx,
                  double xmin, double ymin, double zmin,
                  double eps, int maxit, bool weno, size_t nthreads):
        self.nx = nx
        self.ny = ny
        self.nz = nz
        cdef bool ttrp = 1
        cdef bool interpVel = 0
        self.grid = new Grid3Drnfs[double,uint32_t](nx, ny, nz, dx, xmin, ymin,
                                  zmin, eps, maxit, weno, ttrp, interpVel, nthreads)


    def __dealloc__(self):
        del self.grid

    def get_nthreads(self):
        """
        Returns
        -------
        number of threads allowed for calculations
        """
        return self.grid.getNthreads()

    def set_slowness(self, slowness):
        """
        Assign slowness at grid nodes

        Parameters
        ----------
        slowness : 1D array (in 'C' order)
        """
        cdef vector[double] slown
        nx = self.nx+1
        ny = self.ny+1
        nz = self.nz+1
        for k in range(nz):
            for j in range(ny):
                for i in range(nx):
                    # slowness is in 'C' order and we must pass it in 'F' order
                    slown.push_back(slowness[(i*ny + j)*nz + k])
        self.grid.setSlowness(slown)

    def get_s0(self, Tx, slowness=None):
        """
        Return slowness value at source position

        The source can comprise more than one point

        Parameters
        ----------
            Tx : coordinates of source points (npts x 3)
            slowness : 1D array of slowness values at nodes (can be None, but
                       slowness should have been set before calling get_s0)

        Returns
        -------
            s0 : average slowness at source points
        """
        cdef vector[double] slown
        if slowness is not None:
            nx = self.nx+1
            ny = self.ny+1
            nz = self.nz+1
            for k in range(nz):
                for j in range(ny):
                    for i in range(nx):
                        # slowness is in 'C' order and we must pass it in 'F' order
                        slown.push_back(slowness[(i*ny + j)*nz + k])
            self.grid.setSlowness(slown)

        cdef double s0 = 0.0
        for t in Tx:
            s0 += self.grid.computeSlowness(sxyz[double](t[0], t[1], t[2]))
        return s0 / Tx.shape[0]

    def raytrace(self, slowness, Tx, Rx, t0=0.0, nout=1, thread_no=0):
        """
        Perform raytracing for a single source

        The source can comprise more than one point

        Parameters
        ----------
            slowness : 1D array of slowness values at nodes (can be None, but
                       slowness should have been set before calling raytrace)
            Tx : coordinates of source points (npts x 3)
            Rx : coordinates of receivers (nrcv x 3)
            t0 : time of source event
            nout : number of parameters to return (see below)
            thread_no : thread/process number on which computation should be run

        Returns
        -------
            tt : travel time are receivers
            if nout == 2 or nout == 3:
                rays : list holding coordinates of ray segments, for each rcv
            if nout == 3:
                M : matrix of partial derivatives of t w/r to velocity
        """
        # assing model data
        cdef vector[double] slown
        if slowness is not None:
            nx = self.nx+1
            ny = self.ny+1
            nz = self.nz+1
            for k in range(nz):
                for j in range(ny):
                    for i in range(nx):
                        # slowness is in 'C' order and we must pass it in 'F' order
                        slown.push_back(slowness[(i*ny + j)*nz + k])
            self.grid.setSlowness(slown)

        cdef vector[sxyz[double]] vTx
        cdef vector[sxyz[double]] vRx
        cdef vector[double] vt0
        cdef vector[double] vtt

        for t in Tx:
            vTx.push_back(sxyz[double](t[0], t[1], t[2]))
        for r in Rx:
            vRx.push_back(sxyz[double](r[0], r[1], r[2]))
        vt0.push_back(t0)
        vtt.resize(Rx.shape[0])

        cdef vector[vector[sxyz[double]]] r_data
        cdef vector[vector[sijv[double]]] m_data

        if nout == 1:
            self.grid.raytrace(vTx, vt0, vRx, vtt, thread_no)

            tt = np.empty((Rx.shape[0],))
            for n in range(Rx.shape[0]):
                tt[n] = vtt[n]

            return tt

        elif nout == 2:
            self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, thread_no)

            rays = [ [0.0] for i in range(Rx.shape[0]) ]
            tt = np.empty((Rx.shape[0],))
            for n in range(Rx.shape[0]):
                tt[n] = vtt[n]
                rays[n] = np.empty((r_data[n].size(), 3))
                for nn in range(r_data[n].size()):
                    rays[n][nn, 0] = r_data[n][nn].x
                    rays[n][nn, 1] = r_data[n][nn].y
                    rays[n][nn, 2] = r_data[n][nn].z

            return tt, rays

        elif nout == 3:
            self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data, thread_no)

            rays = [ [0.0] for i in range(Rx.shape[0]) ]
            tt = np.empty((Rx.shape[0],))
            for n in range(Rx.shape[0]):
                tt[n] = vtt[n]
                rays[n] = np.empty((r_data[n].size(), 3))
                for nn in range(r_data[n].size()):
                    rays[n][nn, 0] = r_data[n][nn].x
                    rays[n][nn, 1] = r_data[n][nn].y
                    rays[n][nn, 2] = r_data[n][nn].z

            nnz = 0
            for ni in range(m_data.size()):
                nnz += m_data[ni].size()
            indptr = np.empty((Rx.shape[0]+1,), dtype=np.int64)
            indices = np.empty((nnz,), dtype=np.int64)
            val = np.empty((nnz,))

            k = 0
            M = Rx.shape[0]
            N = (self.nx+1)*(self.ny+1)*(self.nz+1)
            for i in range(M):
                indptr[i] = k
                for j in range(N):
                    for n in range(m_data[i].size()):
                        if m_data[i][n].i == i and m_data[i][n].j == j:
                            indices[k] = j
                            val[k] = m_data[i][n].v
                            k += 1

            indptr[M] = k
            MM = csr_matrix((val, indices, indptr), shape=(M,N))

            return tt, rays, MM


cdef class Grid3Drc:
    """
    Grid3Drc(nx, ny, nz, dx, xmin, ymin, zmin, eps, maxit, weno, nthreads)

    3D rectilinear grid with slowness defined in cells

    Parameters
    ----------
    nx : number of cells in x
    ny : number of cells in y
    nz : number of cells in z
    dx : cell size in x
    xmin : origin in x
    ymin : origin in y
    zmin : origin in z
    eps : convergence criterion (FSM)
    maxit : max number of sweeping iterations
    weno : use 3rd order weighted essentially non-oscillatory operator (bool)
    nthreads : number of threads for raytracing
    """
    cdef uint32_t nx
    cdef uint32_t ny
    cdef uint32_t nz
    cdef Grid3Drcfs[double, uint32_t]* grid
    def __cinit__(self, uint32_t nx, uint32_t ny, uint32_t nz, double dx,
                  double xmin, double ymin, double zmin,
                  double eps, int maxit, bool weno, size_t nthreads):
        self.nx = nx
        self.ny = ny
        self.nz = nz
        cdef bool ttrp = 1
        cdef bool interp_vel = 0
        self.grid = new Grid3Drcfs[double,uint32_t](nx, ny, nz, dx, xmin, ymin,
                                  zmin, eps, maxit, weno, ttrp, interp_vel,
                                  nthreads)

    def __dealloc__(self):
        del self.grid

    def get_nthreads(self):
        """
        Returns
        -------
        number of threads allowed for calculations
        """
        return self.grid.getNthreads()

    def set_slowness(self, slowness):
        """
        Assign slowness at grid cells

        Parameters
        ----------
        slowness : 1D array (in 'C' order)
        """
        cdef vector[double] slown
        for k in range(self.nz):
            for j in range(self.ny):
                for i in range(self.nx):
                    # slowness is in 'C' order and we must pass it in 'F' order
                    slown.push_back(slowness[(i*self.ny + j)*self.nz + k])
        self.grid.setSlowness(slown)

    def raytrace(self, slowness, Tx, Rx, t0, nout=None, thread_no=0, L=None, rays=None):
        """
        Perform raytracing for a single source

        The source can comprise more than one point

        Parameters
        ----------
            slowness : 1D array of slowness values in cells (can be None, but
                       slowness should have been set before calling raytrace)
            Tx : coordinates of source points (npts x 3)
            Rx : coordinates of receivers (nrcv x 3)
            t0 : time of source event
            nout : number of parameters to return (optional, see below)
            thread_no : thread/process number on which computation should be run
            L : ray projection matrix (must be a csr_matrix)
            rays : list holding coordinates of ray segments, for each rcv

        Returns
        -------
            tt : travel time are receivers
            if nout >= 2
                rays : list holding coordinates of ray segments, for each rcv
            if nout == 3:
                L : ray projection matrix

        Note
        ----
            if nout is given, input arguments L and rays are ignored, new intances
            are created if needed
        """
        # assing model data
        cdef vector[double] slown
        if slowness is not None:
            for k in range(self.nz):
                for j in range(self.ny):
                    for i in range(self.nx):
                        # slowness is in 'C' order and we must pass it in 'F' order
                        slown.push_back(slowness[(i*self.ny + j)*self.nz + k])
            self.grid.setSlowness(slown)

        cdef vector[sxyz[double]] vTx
        cdef vector[sxyz[double]] vRx
        cdef vector[double] vt0
        cdef vector[double] vtt

        for t in Tx:
            vTx.push_back(sxyz[double](t[0], t[1], t[2]))
        for r in Rx:
            vRx.push_back(sxyz[double](r[0], r[1], r[2]))
        vt0.push_back(t0)
        vtt.resize(Rx.shape[0])

        cdef vector[vector[siv[double]]] l_data
        cdef vector[vector[sxyz[double]]] r_data
        cdef double s0 = 0.0

        if nout is not None:
            if nout == 1:
                self.grid.raytrace(vTx, vt0, vRx, vtt, thread_no)

                tt = np.empty((Rx.shape[0],))
                for n in range(Rx.shape[0]):
                    tt[n] = vtt[n]
                return tt
            elif nout == 2:
                self.grid.raytrace2(vTx, vt0, vRx, vtt, r_data, thread_no)

                rays = [ [0.0] for i in range(Rx.shape[0]) ]
                tt = np.empty((Rx.shape[0],))
                for n in range(Rx.shape[0]):
                    tt[n] = vtt[n]
                    rays[n] = np.empty((r_data[n].size(), 3))
                    for nn in range(r_data[n].size()):
                        rays[n][nn, 0] = r_data[n][nn].x
                        rays[n][nn, 1] = r_data[n][nn].y
                        rays[n][nn, 2] = r_data[n][nn].z

                return tt, rays
            elif nout == 3:
                self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data, thread_no)

                rays = [ [0.0] for i in range(Rx.shape[0]) ]
                tt = np.empty((Rx.shape[0],))
                for n in range(Rx.shape[0]):
                    tt[n] = vtt[n]
                    rays[n] = np.empty((r_data[n].size(), 3))
                    for nn in range(r_data[n].size()):
                        rays[n][nn, 0] = r_data[n][nn].x
                        rays[n][nn, 1] = r_data[n][nn].y
                        rays[n][nn, 2] = r_data[n][nn].z

                indptr = np.empty((Rx.shape[0]+1,), dtype=np.int64)
                nval = 0
                for i in range(l_data.size()):
                    nval += l_data[i].size()
                indices = np.empty((nval,), dtype=np.int64)
                val = np.empty((nval,))

                k = 0
                M = Rx.shape[0]
                N = self.nx*self.ny*self.nz
                for i in range(M):
                    indptr[i] = k
                    for j in range(N):
                        for n in range(l_data[i].size()):
                            if l_data[i][n].i == j:
                                indices[k] = j
                                val[k] = l_data[i][n].v
                                k += 1

                indptr[M] = k
                L = csr_matrix((val, indices, indptr), shape=(M,N))

                return tt, rays, L
        else:

            if L is None and rays is None:
                self.grid.raytrace(vTx, vt0, vRx, vtt, thread_no)

                tt = np.empty((Rx.shape[0],))
                for n in range(Rx.shape[0]):
                    tt[n] = vtt[n]

            elif L is not None and rays is None:

                if type(L) is not csr_matrix:
                    raise TypeError('L should be csr_matrix')

                self.grid.raytrace(vTx, vt0, vRx, vtt, l_data, thread_no)

                tt = np.empty((Rx.shape[0],))
                for n in range(Rx.shape[0]):
                    tt[n] = vtt[n]

                indptr = np.empty((Rx.shape[0]+1,), dtype=np.int64)
                nval = 0
                for i in range(l_data.size()):
                    nval += l_data[i].size()
                indices = np.empty((nval,), dtype=np.int64)
                val = np.empty((nval,))

                k = 0
                M = Rx.shape[0]
                N = self.nx*self.ny*self.nz
                for i in range(M):
                    indptr[i] = k
                    for j in range(N):
                        for n in range(l_data[i].size()):
                            if l_data[i][n].i == j:
                                indices[k] = j
                                val[k] = l_data[i][n].v
                                k += 1

                indptr[M] = k

                L.data = val
                L.indices = indices
                L.indptr = indptr
                L.reshape((M,N))

            elif L is None and rays is not None:
                self.grid.raytrace2(vTx, vt0, vRx, vtt, r_data, thread_no)

                rays.clear()
                tt = np.empty((Rx.shape[0],))
                for n in range(Rx.shape[0]):
                    tt[n] = vtt[n]
                    rays.append(np.empty((r_data[n].size(), 3)))
                    for nn in range(r_data[n].size()):
                        rays[-1][nn, 0] = r_data[n][nn].x
                        rays[-1][nn, 1] = r_data[n][nn].y
                        rays[-1][nn, 2] = r_data[n][nn].z

            elif L is not None and rays is not None:

                if type(L) is not csr_matrix:
                    raise TypeError('L should be csr_matrix')

                self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data, thread_no)

                rays.clear()
                tt = np.empty((Rx.shape[0],))
                for n in range(Rx.shape[0]):
                    tt[n] = vtt[n]
                    rays.append(np.empty((r_data[n].size(), 3)))
                    for nn in range(r_data[n].size()):
                        rays[-1][nn, 0] = r_data[n][nn].x
                        rays[-1][nn, 1] = r_data[n][nn].y
                        rays[-1][nn, 2] = r_data[n][nn].z

                indptr = np.empty((Rx.shape[0]+1,), dtype=np.int64)
                nval = 0
                for i in range(l_data.size()):
                    nval += l_data[i].size()
                indices = np.empty((nval,), dtype=np.int64)
                val = np.empty((nval,))

                k = 0
                M = Rx.shape[0]
                N = self.nx*self.ny*self.nz
                for i in range(M):
                    indptr[i] = k
                    for j in range(N):
                        for n in range(l_data[i].size()):
                            if l_data[i][n].i == j:
                                indices[k] = j
                                val[k] = l_data[i][n].v
                                k += 1

                indptr[M] = k

                L.data = val
                L.indices = indices
                L.indptr = indptr
                L.reshape((M,N))

            return tt



# -*- coding: utf-8 -*-
"""
.. deprecated:: v0.4.0
    `cmesh2d` will be removed in ttcrpy v1.0.0, it is replaced by
    `tmesh`.

Code for raytracing on 2D triangular meshes based on the shortest path method

This code is part of ttcr ( https://github.com/groupeLIAMG/ttcr )
"""

"""
Copyright 2018 Bernard Giroux
email: bernard.giroux@ete.inrs.ca

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from libcpp.string cimport string
from libcpp.vector cimport vector
from libc.stdint cimport uint32_t

import numpy as np
cimport numpy as np

from scipy.sparse import csr_matrix

import vtk
from vtk.util import numpy_support

cdef extern from "ttcr_t.h" namespace "ttcr":
    cdef cppclass sxz[T]:
        sxz(T, T) except +
        T x
        T z

    cdef cppclass siv[T]:
        siv(size_t, T) except +
        size_t i
        T v

    cdef cppclass triangleElem[T]:
        triangleElem(T, T, T) except +
        T i[3]
        T physical_entity


cdef extern from "Node2Dcsp.h" namespace "ttcr":
    cdef cppclass Node2Dcsp[T1,T2]:
        Node2Dcsp(size_t) except +


cdef extern from "Grid2Ducsp.h" namespace "ttcr":
    cdef cppclass Grid2Ducsp[T1,T2,NODE,S]:
        Grid2Ducsp(vector[S], vector[triangleElem[T2]], T2, size_t) except +
        size_t getNthreads()
        size_t getNumberOfCells()
        size_t getNumberOfNodes(bool)
        void setSlowness(T1 *, size_t) except +
        void raytrace(vector[S]&, vector[T1]&, vector[S]&, vector[T1]&, size_t) except +
        void raytrace(vector[S]&, vector[T1]&, vector[S]&, vector[T1]&, vector[vector[S]]&, vector[vector[siv[T1]]]&, size_t) except +
        void calculateArea(vector[T1] &) except +
        void interpolateAtNodes(vector[T1] &) except +
        void getTT(vector[T1]&, size_t)
        void getNodes(vector[S]&)
        void getTriangles(vector[vector[T2]]&)


cdef class Mesh2D:
    """
        Wrapper class for the C++ code

        Constructor

            Mesh2D(nodes, triangles, nsecondary, nthreads)

            Parameters
            ----------
            nodes : coordinates of nodes (ndarray of size nnodes x 2, double)
                   1st column contains X coordinates,
                   2nd contains Z coordinates
            triangles : indices of nodes forming triangles (ndarray of size ntriange x 3, int)
            nsecondary : number of secondary nodes in triangle edges (int)
            nthreads : number of threads/processes for raytracing (int)
    """
    cdef Grid2Ducsp[double, uint32_t, Node2Dcsp[double, uint32_t], sxz[double]] *mesh
    def __cinit__(self, nodes, triangles, uint32_t nsecondary, size_t nthreads):
        cdef vector[sxz[double]] no
        for n in nodes:
            no.push_back(sxz[double](n[0], n[1]))
        cdef vector[triangleElem[uint32_t]] tri
        for t in triangles:
            tri.push_back(triangleElem[uint32_t](t[0], t[1], t[2]))
        self.mesh = new Grid2Ducsp[double, uint32_t, Node2Dcsp[double, uint32_t], sxz[double]](no, tri, nsecondary, nthreads)

    def __dealloc__(self):
        del self.mesh

    def get_grid_traveltimes(self, thread_no=0):
        """
        Obtain traveltimes computed at primary grid nodes

        Parameters
        ----------
        thread_no : int
            thread used to computed traveltimes (default is 0)

        Returns
        -------
        tt: np ndarray
        """
        if thread_no >= self.mesh.getNthreads():
            raise ValueError('Thread number is larger than number of threads')
        cdef vector[double] tmp
        cdef int n
        self.mesh.getTT(tmp, thread_no)
        tt = np.empty((tmp.size(),))
        for n in range(tmp.size()):
            tt[n] = tmp[n]
        return tt

    def set_slowness(self, slowness):
        """
        Assign slowness at cells

        set_slowness(slowness)

        Parameters
        ----------
        slowness : ndarray of size ntriangle.  Indices of slowness values
                   must correspond to triangle indices given in constructor
        """
        if not slowness.flags['C_CONTIGUOUS']:
            slowness = np.ascontiguousarray(slowness)

        # ::1 tells cython that data is contiguous
        cdef double[::1] slowness_memview = slowness

        self.mesh.setSlowness(&slowness_memview[0], slowness_memview.shape[0])

    def raytrace(self, slowness, Tx, Rx, t0=0.0, nout=1, thread_no=0):
        """
        raytrace(slowness, Tx, Rx, t0, nout, thread_no) -> tt,L,rays

        Compute traveltimes for a single source.  The source can comprise many
        points, each having its own otigin time (t0).

        Parameters
        ----------
        slowness : vector of slowness at cells
        Tx : source coordinates, ndarray of size nTx by 2
            1st column contains X coordinates,
            2nd contains Z coordinates
        Rx : receiver coordinates, ndarray of size nRx by 2
            1st column contains X coordinates,
            2nd contains Z coordinates
        t0 : origin times, scalar or ndarray of size nTx
        nout : number of output variables (1, 2 or 3)
        thread_no : thread/process number on which computation should be run

        Returns
        -------
        tt : travel times
        L : (if nout > 1) data kernel matrix (scipy csr) (tt = L*slowness)
        rays : (if nout == 3) coordinates of rays
        """
        cdef double[::1] slowness_memview
        if slowness is not None:
            if not slowness.flags['C_CONTIGUOUS']:
                slowness = np.ascontiguousarray(slowness)

            slowness_memview = slowness
            self.mesh.setSlowness(&slowness_memview[0], slowness_memview.shape[0])

        cdef vector[sxz[double]] vTx
        cdef vector[sxz[double]] vRx
        cdef vector[double] vt0
        cdef vector[double] vtt

        for t in Tx:
            vTx.push_back(sxz[double](t[0], t[1]))
        for r in Rx:
            vRx.push_back(sxz[double](r[0], r[1]))

        if np.isscalar(t0):
            for n in range(Tx.shape[0]):
                vt0.push_back(t0)
        else:
            for t in t0:
                vt0.push_back(t)

        vtt.resize(Rx.shape[0])

        cdef vector[vector[sxz[double]]] r_data
        cdef vector[vector[siv[double]]] l_data
        cdef double v0 = 0.0

        if nout == 1:
            self.mesh.raytrace(vTx, vt0, vRx, vtt, thread_no)

            tt = np.empty((Rx.shape[0],))
            for n in range(Rx.shape[0]):
                tt[n] = vtt[n]

            return tt

        elif nout==2:
            self.mesh.raytrace(vTx, vt0, vRx, vtt, r_data, l_data, thread_no)

            tt = np.empty((Rx.shape[0],))
            for n in range(Rx.shape[0]):
                tt[n] = vtt[n]

            indptr = np.empty((Rx.shape[0]+1,), dtype=np.int64)
            nel = 0
            for n in range(l_data.size()):
                nel += l_data[n].size()
            indices = np.empty((nel,), dtype=np.int64)
            val = np.empty((nel,))

            k = 0
            M = Rx.shape[0]
            N = self.mesh.getNumberOfCells()
            for i in range(M):
                indptr[i] = k
                for j in range(N):
                    for n in range(l_data[i].size()):
                        if l_data[i][n].i == j:
                            indices[k] = j
                            val[k] = l_data[i][n].v
                            k += 1

            indptr[M] = k
            L = csr_matrix((val, indices, indptr), shape=(M,N))

            return tt, L

        elif nout == 3:
            self.mesh.raytrace(vTx, vt0, vRx, vtt, r_data, l_data, thread_no)

            rays = [ [0.0] for i in range(Rx.shape[0]) ]
            tt = np.empty((Rx.shape[0],))
            for n in range(Rx.shape[0]):
                tt[n] = vtt[n]
                rays[n] = np.empty((r_data[n].size(), 2))
                for nn in range(r_data[n].size()):
                    rays[n][nn, 0] = r_data[n][nn].x
                    rays[n][nn, 1] = r_data[n][nn].z

            indptr = np.empty((Rx.shape[0]+1,), dtype=np.int64)
            nel = 0
            for n in range(l_data.size()):
                nel += l_data[n].size()
            indices = np.empty((nel,), dtype=np.int64)
            val = np.empty((nel,))

            k = 0
            M = Rx.shape[0]
            N = self.mesh.getNumberOfCells()
            for i in range(M):
                indptr[i] = k
                for j in range(N):
                    for n in range(l_data[i].size()):
                        if l_data[i][n].i == j:
                            indices[k] = j
                            val[k] = l_data[i][n].v
                            k += 1

            indptr[M] = k
            L = csr_matrix((val, indices, indptr), shape=(M,N))

            return tt, L, rays

    def to_vtk(self, fields, filename):
        """
        Save grid variables to VTK format

        Parameters
        ----------
        fields: dict
            dict of variables to save to file
            variables should be numpy ndarrays of size equal to either the
            number of nodes of the number of cells of the grid
        filename: str
            Name of file without extension for saving (extension vtr will be
            added)

        Notes
        -----
        VTK files can be visualized with Paraview (https://www.paraview.org)
        """
        cdef vector[sxz[double]] nodes
        cdef vector[vector[uint32_t]] triangles
        self.mesh.getNodes(nodes)
        self.mesh.getTriangles(triangles)

        cdef int n
        ugrid = vtk.vtkUnstructuredGrid()
        pts = vtk.vtkPoints()
        tri = vtk.vtkTriangle()
        for n in range(nodes.size()):
            pts.InsertNextPoint(nodes[n].x, 0.0, nodes[n].z)
        ugrid.SetPoints(pts)
        for n in range(triangles.size()):
            tri.GetPointIds().SetId(0, triangles[n][0])
            tri.GetPointIds().SetId(1, triangles[n][1])
            tri.GetPointIds().SetId(2, triangles[n][2])
            ugrid.InsertNextCell(tri.GetCellType(), tri.GetPointIds())

        for fn in fields:
            data = fields[fn]

            scalar = vtk.vtkDoubleArray()
            scalar.SetName(fn)
            scalar.SetNumberOfComponents(1)
            scalar.SetNumberOfTuples(data.size)
            if data.size == self.mesh.getNumberOfNodes(1):
                for n in range(nodes.size()):
                    scalar.SetTuple1(n, data[n])
                ugrid.GetPointData().AddArray(scalar)
            elif data.size == self.mesh.getNumberOfCells():
                for n in range(triangles.size()):
                    scalar.SetTuple1(n, data[n])
                ugrid.GetCellData().AddArray(scalar)
            else:
                raise ValueError('Field {0:s} has incorrect size'.format(fn))

            writer = vtk.vtkXMLUnstructuredGridWriter()
            writer.SetFileName(filename+'.vtu')
            writer.SetInputData(ugrid)
            writer.SetDataModeToBinary()
            writer.Update()

#    def get_Dx_Dz(self):
#        """
#            Compute spatial derivatives on mesh
#        """
#        cdef vector[double] area
#        self.mesh.calculateArea(area)
#
#        # TODO: complete



# -*- coding: utf-8 -*-
"""Tests for verifying python wrappers, module rgrid in 3D"""

import unittest
import numpy as np
import vtk
from vtk.util.numpy_support import vtk_to_numpy

vtk.vtkObject.GlobalWarningDisplayOff()

import ttcrpy.rgrid as rg


def get_tt(filename):
    reader = vtk.vtkXMLRectilinearGridReader()
    reader.SetFileName(filename)
    reader.Update()
    data = reader.GetOutput()
    names = ('Travel Time', 'Travel time', 'travel time')
    for name in names:
        if data.GetPointData().HasArray(name):
            break
    tt = vtk_to_numpy(data.GetPointData().GetArray(name))
    return tt


class TestGrid3dc(unittest.TestCase):

    def setUp(self):
        reader = vtk.vtkXMLRectilinearGridReader()
        reader.SetFileName('layers_coarse.vtr')
        reader.Update()

        data = reader.GetOutput()
        self.x = vtk_to_numpy(data.GetXCoordinates())
        self.y = vtk_to_numpy(data.GetYCoordinates())
        self.z = vtk_to_numpy(data.GetZCoordinates())

        self.slowness = vtk_to_numpy(data.GetCellData().GetArray('Slowness'))
        dim = (self.x.size-1, self.y.size-1, self.z.size-1)
        self.slowness = self.slowness.reshape(dim, order='F').flatten()
        self.src = np.loadtxt('src.dat',skiprows=1)
        self.src = self.src.reshape((1, 4))
        self.rcv = np.loadtxt('rcv.dat',skiprows=1)

    def test_Grid3Dfs(self):
        g = rg.Grid3d(self.x, self.y, self.z, method='FSM', tt_from_rp=False)
        tt = g.raytrace(self.src, self.rcv, self.slowness)
        dim = (self.x.size, self.y.size, self.z.size)
        tt = g.get_grid_traveltimes()
        tt = tt.flatten()
        tt_ref = get_tt('fsm_d_p_lc_src_all_tt.vtr')
        self.assertLess(np.sum(np.abs(tt-tt_ref))/tt.size, 0.01,
                        'FSM accuracy failed (slowness in cells)')

    def test_Grid3Dsp(self):
        g = rg.Grid3d(self.x, self.y, self.z, method='SPM', tt_from_rp=False,
                      nsnx=5, nsny=5, nsnz=5)
        tt = g.raytrace(self.src, self.rcv, self.slowness)
        dim = (self.x.size, self.y.size, self.z.size)
        tt = g.get_grid_traveltimes()
        tt = tt.flatten()
        tt_ref = get_tt('spm_d_p_lc_05_src_all_tt.vtr')
        self.assertLess(np.sum(np.abs(tt-tt_ref))/tt.size, 0.1,
                        'SPM accuracy failed (slowness in cells)')

    def test_Grid3Ddsp(self):
        g = rg.Grid3d(self.x, self.y, self.z, method='DSPM', tt_from_rp=False,
                      n_secondary=2, n_tertiary=2, radius_tertiary=2)
        tt = g.raytrace(self.src, self.rcv, self.slowness)
        dim = (self.x.size, self.y.size, self.z.size)
        tt = g.get_grid_traveltimes()
        tt = tt.flatten()
        tt_ref = get_tt('dspm_d_p_lc_2_2_2_src_all_tt.vtr')
        self.assertLess(np.sum(np.abs(tt-tt_ref))/tt.size, 0.1,
                        'SPM accuracy failed (slowness in cells)')


class TestGrid3dn(unittest.TestCase):

    def setUp(self):
        reader = vtk.vtkXMLRectilinearGridReader()
        reader.SetFileName('gradient_coarse.vtr')
        reader.Update()

        data = reader.GetOutput()
        self.x = vtk_to_numpy(data.GetXCoordinates())
        self.y = vtk_to_numpy(data.GetYCoordinates())
        self.z = vtk_to_numpy(data.GetZCoordinates())

        self.slowness = vtk_to_numpy(data.GetPointData().GetArray('Slowness'))
        dim = (self.x.size, self.y.size, self.z.size)
        self.slowness = self.slowness.reshape(dim, order='F').flatten()
        self.src = np.loadtxt('src.dat',skiprows=1)
        self.src = self.src.reshape((1, 4))
        self.rcv = np.loadtxt('rcv.dat',skiprows=1)

    def test_Grid3Dfs(self):
        g = rg.Grid3d(self.x, self.y, self.z, method='FSM', tt_from_rp=False,
                      cell_slowness=0)
        tt = g.raytrace(self.src, self.rcv, self.slowness)
        dim = (self.x.size, self.y.size, self.z.size)
        tt = g.get_grid_traveltimes()
        tt = tt.flatten()
        tt_ref = get_tt('fsm_d_p_gc_src_all_tt.vtr')
        self.assertLess(np.sum(np.abs(tt-tt_ref))/tt.size, 0.01,
                        'FSM accuracy failed (slowness at nodes)')

    def test_Grid3Dsp(self):
        g = rg.Grid3d(self.x, self.y, self.z, method='SPM', tt_from_rp=False,
                      nsnx=5, nsny=5, nsnz=5, cell_slowness=0)
        tt = g.raytrace(self.src, self.rcv, self.slowness)
        dim = (self.x.size, self.y.size, self.z.size)
        tt = g.get_grid_traveltimes()
        tt = tt.flatten()
        tt_ref = get_tt('spm_d_p_gc_05_src_all_tt.vtr')
        self.assertLess(np.sum(np.abs(tt-tt_ref))/tt.size, 0.1,
                        'SPM accuracy failed (slowness at nodes)')

    def test_Grid3Ddsp(self):
        g = rg.Grid3d(self.x, self.y, self.z, method='DSPM', tt_from_rp=False,
                      n_secondary=2, n_tertiary=2, radius_tertiary=2,
                      cell_slowness=0)
        tt = g.raytrace(self.src, self.rcv, self.slowness)
        dim = (self.x.size, self.y.size, self.z.size)
        tt = g.get_grid_traveltimes()
        tt = tt.flatten()
        tt_ref = get_tt('dspm_d_p_gc_2_2_2_src_all_tt.vtr')
        self.assertLess(np.sum(np.abs(tt-tt_ref))/tt.size, 0.1,
                        'SPM accuracy failed (slowness at nodes)')


class Data_kernel(unittest.TestCase):

    def test_3d(self):

        V = np.ones((11, 12, 13))
        V[:, :, 7:] = 2
        slowness = 1. / V.flatten()

        grx = np.arange(12.)
        gry = np.arange(13.)
        grz = np.arange(14.)

        z = 0.5 + np.arange(13.)
        Tx = np.vstack((0.5+np.zeros((13,)),
                        0.5+np.zeros((13,)),
                        z)).T
        Rx = np.vstack((10.5+np.zeros((13,)),
                        11.5+np.zeros((13,)),
                        z)).T
        nTx = Tx.shape[0]
        nRx = Rx.shape[0]
        Tx = np.kron(Tx, np.ones((nRx,1)))
        Rx = np.kron(np.ones((nTx,1)), Rx)

        L = rg.Grid3d.data_kernel_straight_rays(Tx, Rx, grx, gry, grz)
        tt = L.dot(slowness)

        tt2 = np.zeros(tt.shape)
        d = np.sqrt(np.sum((Tx-Rx)**2, axis=1))

        ind = np.logical_and(Tx[:,2]>7, Rx[:,2]>7)
        tt2[ind] = d[ind]/2

        ind2 = np.logical_and(Tx[:,2]<7, Rx[:,2]<7)
        tt2[ind2] = d[ind2]

        ind3 = np.logical_and(np.logical_not(ind), np.logical_not(ind2))

        f = (7-Tx[ind3,2]) / (Rx[ind3,2]-Tx[ind3,2])
        ind = (Rx[ind3,2]-Tx[ind3,2]) < 0
        f[ind] = 1-f[ind]
        tt2[ind3] = d[ind3]*f + d[ind3]*(1-f)/2

        self.assertAlmostEqual(np.sum(np.abs(tt-tt2)), 0.0 )

if __name__ == '__main__':

    unittest.main()



# -*- coding: utf-8 -*-
"""
.. deprecated:: v0.4.0
    `cmesh3d` will be removed in ttcrpy v1.0.0, it is replaced by
    `tmesh`.

Code for raytracing on 3D tetrahedral meshes based on the shortest path method

This code is part of ttcr ( https://github.com/groupeLIAMG/ttcr )
"""

"""
Copyright 2016 Bernard Giroux
email: bernard.giroux@ete.inrs.ca

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


from libcpp.string cimport string
from libcpp.vector cimport vector
from libc.stdint cimport uint32_t
from libcpp cimport bool

import numpy as np
cimport numpy as np

from scipy.sparse import csr_matrix

cdef extern from "ttcr_t.h" namespace "ttcr":
    cdef cppclass sxyz[T]:
        sxyz(T, T, T) except +

cdef extern from "ttcr_t.h" namespace "ttcr":
    cdef cppclass tetrahedronElem[T]:
        tetrahedronElem(T, T, T, T) except +


cdef extern from "Mesh3Dttcr.h" namespace "ttcr":
    cdef cppclass Mesh3Dttcr:
        Mesh3Dttcr(vector[sxyz[double]]&, vector[tetrahedronElem[uint32_t]]&,
                   double, int, bool rp, size_t) except +
        void setSlowness(const vector[double]&) except +
        void raytrace(const vector[sxyz[double]]&,
                     const vector[double]&,
                     const vector[sxyz[double]]&,
                     double*) except +
        void raytrace(const vector[sxyz[double]]&,
                     const vector[double]&,
                     const vector[sxyz[double]]&,
                     double*, object, double*) except +
#        void raytrace(const vector[sxyz[double]]& Tx,
#                     const vector[double]& tTx,
#                     const vector[sxyz[double]]& Rx,
#                     double*, object, double*, object) except +

cdef class Mesh3Dcpp:
    cdef Mesh3Dttcr* mesh
    def __cinit__(self, nodes, tetra, const double eps, const int maxit, const bool rp, const size_t nt):
        cdef vector[sxyz[double]] nod
        for no in nodes:
            nod.push_back(sxyz[double](no[0], no[1], no[2]))
        cdef vector[tetrahedronElem[uint32_t]] tet
        for t in tetra:
            tet.push_back(tetrahedronElem[uint32_t](t[0], t[1], t[2], t[3]))
        self.mesh = new Mesh3Dttcr(nod, tet, eps, maxit, rp, nt)

    def __dealloc__(self):
        del self.mesh

    def raytrace(self, slowness, Tx, Rx, t0):
        nout = 4

        # assing model data
        cdef vector[double] slown
        for tmp in slowness:
            slown.push_back(tmp)
        self.mesh.setSlowness(slown)

        # create C++ input variables
        cdef vector[sxyz[double]] cTx
        for t in Tx:
            cTx.push_back(sxyz[double](t[0], t[1], t[2]))

        cdef vector[sxyz[double]] cRx
        for r in Rx:
            cRx.push_back(sxyz[double](r[0], r[1], r[2]))

        cdef vector[double] ct0
        for tmp in t0:
            ct0.push_back(tmp)

        # instantiate output variables
        cdef np.ndarray tt = np.empty([Rx.shape[0],], dtype=np.double)  # tt should be the right size
        cdef np.ndarray v0 = np.empty([Rx.shape[0],], dtype=np.double)

        if nout == 1:
            self.mesh.raytrace(cTx, ct0, cRx, <double*> np.PyArray_DATA(tt))
            return tt

        elif nout == 3:
            rays = tuple([ [0.0] for i in range(Rx.shape[0]) ])

            self.mesh.raytrace(cTx, ct0, cRx, <double*> np.PyArray_DATA(tt), rays, <double*> np.PyArray_DATA(v0))
            return tt, rays, v0

#        elif nout == 4:
#            rays = tuple([ [0.0] for i in range(Rx.shape[0]) ])
#
#            nTx = np.unique(Tx, axis=0).shape[0]
#
#            M = tuple([ ([0.0],[0.0],[0.0]) for i in range(nTx) ])
#
#            self.mesh.raytrace(cTx, ct0, cRx, <double*> np.PyArray_DATA(tt), rays, <double*> np.PyArray_DATA(v0), M)
#            return tt, rays, v0, M




from libcpp.string cimport string
from libcpp.vector cimport vector
from libc.stdint cimport uint32_t, int64_t
from libc.math cimport sqrt
from libcpp cimport bool

from ttcrpy.common cimport sxz, sxyz, siv, siv2, sijv, Node3Dc, Node3Dcsp, \
Node3Dn, Node3Dnsp, Cell, Node2Dc, Node2Dcsp, Node2Dn, Node2Dnsp


cdef extern from "ttcr_t.h" namespace "ttcr" nogil:
    cdef cppclass triangleElem[T]:
        triangleElem(T, T, T) except +
        T i[3]
        T physical_entity

cdef extern from "ttcr_t.h" namespace "ttcr" nogil:
    cdef cppclass tetrahedronElem[T]:
        tetrahedronElem(T, T, T, T) except +
        T i[4]
        T physical_entity


cdef extern from "Grid3D.h" namespace "ttcr" nogil:
    cdef cppclass Grid3D[T1,T2]:
        void setUsePool(bool up)
        void checkPts(vector[sxyz[T1]]&) except +
        void computeD(vector[sxyz[T1]] &, vector[vector[sijv[T1]]] &)
        void computeK(vector[vector[vector[siv[T1]]]]&, int, int, bool, bool) except +
        size_t getNthreads()
        void setSlowness(vector[T1]&) except +
        T1 computeSlowness(sxyz[T1]&) except +
        void getTT(vector[T1]& tt, size_t threadNo) except +
        void raytrace(vector[sxyz[T1]]& Tx,
                      vector[T1]& t0,
                      vector[sxyz[T1]]& Rx,
                      vector[T1]& tt,
                      size_t thread_no) except +
        void raytrace(vector[sxyz[T1]]& Tx,
                      vector[T1]& t0,
                      vector[sxyz[T1]]& Rx,
                      vector[T1]& tt,
                      vector[vector[sxyz[T1]]]& r_data,
                      size_t thread_no) except +
        void raytrace(vector[sxyz[T1]]& Tx,
                      vector[T1]& t0,
                      vector[sxyz[T1]]& Rx,
                      vector[T1]& tt,
                      vector[vector[siv[T1]]]& l_data,
                      size_t thread_no) except +
        void raytrace(vector[sxyz[T1]]& Tx,
                      vector[T1]& t0,
                      vector[sxyz[T1]]& Rx,
                      vector[T1]& tt,
                      vector[vector[sxyz[T1]]]& r_data,
                      vector[vector[siv[T1]]]& l_data,
                      size_t thread_no) except +
        void raytrace(vector[sxyz[T1]]& Tx,
                      vector[T1]& t0,
                      vector[sxyz[T1]]& Rx,
                      vector[T1]& tt,
                      vector[vector[sijv[T1]]]& m_data,
                      size_t thread_no) except +
        void raytrace(vector[sxyz[T1]]& Tx,
                      vector[T1]& t0,
                      vector[sxyz[T1]]& Rx,
                      vector[T1]& tt,
                      vector[vector[sxyz[T1]]]& r_data,
                      vector[vector[sijv[T1]]]& m_data,
                      size_t thread_no) except +
        void raytrace(vector[vector[sxyz[T1]]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[sxyz[T1]]]& Rx,
                      vector[vector[T1]]& traveltimes) except +
        void raytrace(vector[vector[sxyz[T1]]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[sxyz[T1]]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[sxyz[T1]]]]& r_data) except +
        void raytrace(vector[vector[sxyz[T1]]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[sxyz[T1]]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[sijv[T1]]]]& m_data) except +
        void raytrace(vector[vector[sxyz[T1]]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[sxyz[T1]]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[sxyz[T1]]]]& r_data,
                      vector[vector[vector[sijv[T1]]]]& m_data) except +
        void raytrace(vector[vector[sxyz[T1]]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[sxyz[T1]]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[siv[T1]]]]& l_data) except +
        void raytrace(vector[vector[sxyz[T1]]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[sxyz[T1]]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[sxyz[T1]]]]& r_data,
                      vector[vector[vector[siv[T1]]]]& l_data) except +

cdef extern from "Grid3Duc.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Duc[T1,T2,N](Grid3D[T1,T2]):
        pass

cdef extern from "Grid3Dun.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Dun[T1,T2,N](Grid3D[T1,T2]):
        pass

cdef extern from "Grid3Ducfs.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Ducfs[T1, T2](Grid3Duc[T1,T2,Node3Dc[T1,T2]]):
        Grid3Ducfs(vector[sxyz[T1]], vector[tetrahedronElem[T2]], T1, int,
                   vector[sxyz[T1]]& ref_pts, int order, bool, bool, T1,
                   size_t) except +

cdef extern from "Grid3Ducsp.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Ducsp[T1, T2](Grid3Duc[T1,T2,Node3Dcsp[T1,T2]]):
        Grid3Ducsp(vector[sxyz[T1]], vector[tetrahedronElem[T2]],
                   int, bool, T1, size_t) except +

cdef extern from "Grid3Ducdsp.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Ducdsp[T1, T2](Grid3Duc[T1,T2,Node3Dc[T1,T2]]):
        Grid3Ducdsp(vector[sxyz[T1]], vector[tetrahedronElem[T2]],
                    int, int, T1, int, bool, T1, T1, size_t) except +

cdef extern from "Grid3Dunfs.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Dunfs[T1, T2](Grid3Dun[T1,T2,Node3Dn[T1,T2]]):
        Grid3Dunfs(vector[sxyz[T1]], vector[tetrahedronElem[T2]], T1, int,
                   vector[sxyz[T1]]& ref_pts, int order, int, bool, bool, T1,
                   size_t) except +

cdef extern from "Grid3Dunsp.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Dunsp[T1, T2](Grid3Dun[T1,T2,Node3Dnsp[T1,T2]]):
        Grid3Dunsp(vector[sxyz[T1]], vector[tetrahedronElem[T2]], int, bool,
                   bool, T1, size_t) except +

cdef extern from "Grid3Dundsp.h" namespace "ttcr" nogil:
    cdef cppclass Grid3Dundsp[T1, T2](Grid3Dun[T1,T2,Node3Dn[T1,T2]]):
        Grid3Dundsp(vector[sxyz[T1]], vector[tetrahedronElem[T2]], int, int, T1,
                    bool, int, bool, T1, T1, size_t) except +


cdef extern from "Grid2D.h" namespace "ttcr" nogil:
    cdef cppclass Grid2D[T1,T2,S]:
        size_t getNthreads()
        void setSlowness(vector[T1]&) except +
        void getSlowness(vector[T1]&) except +
        void getTT(vector[T1]& tt, size_t threadNo) except +
        void raytrace(vector[S]& Tx,
                      vector[T1]& t0,
                      vector[S]& Rx,
                      vector[T1]& traveltimes,
                      size_t threadNo) except +
        void raytrace(vector[S]& Tx,
                      vector[T1]& t0,
                      vector[S]& Rx,
                      vector[T1]& traveltimes,
                      vector[vector[sxz[T1]]]& r_data,
                      size_t threadNo) except +
        void raytrace(vector[S]& Tx,
                      vector[T1]& t0,
                      vector[S]& Rx,
                      vector[T1]& traveltimes,
                      vector[vector[siv2[T1]]]& l_data,
                      size_t threadNo) except +
        void raytrace(vector[S]& Tx,
                      vector[T1]& t0,
                      vector[S]& Rx,
                      vector[T1]& traveltimes,
                      vector[vector[S]]& r_data,
                      vector[vector[siv2[T1]]]& l_data,
                      size_t threadNo) except +
        void raytrace(vector[vector[S]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[S]]& Rx,
                      vector[vector[T1]]& traveltimes) except +
        void raytrace(vector[vector[S]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[S]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[S]]]& r_data) except +
        void raytrace(vector[vector[S]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[S]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[siv2[T1]]]]& l_data) except +
        void raytrace(vector[vector[S]]& Tx,
                      vector[vector[T1]]& t0,
                      vector[vector[S]]& Rx,
                      vector[vector[T1]]& traveltimes,
                      vector[vector[vector[S]]]& r_data,
                      vector[vector[vector[siv2[T1]]]]& l_data) except +

cdef extern from "Grid2Duc.h" namespace "ttcr" nogil:
    cdef cppclass Grid2Duc[T1,T2,NODE,S](Grid2D[T1,T2,S]):
        pass

cdef extern from "Grid2Dun.h" namespace "ttcr" nogil:
    cdef cppclass Grid2Dun[T1,T2,NODE,S](Grid2D[T1,T2,S]):
        pass

cdef extern from "Grid2Ducsp.h" namespace "ttcr" nogil:
    cdef cppclass Grid2Ducsp[T1,T2,NODE,S](Grid2Duc[T1,T2,NODE,S]):
        Grid2Ducsp(vector[S]&, vector[triangleElem[T2]]&, T2, size_t) except +

cdef extern from "Grid2Ducfs.h" namespace "ttcr" nogil:
    cdef cppclass Grid2Ducfs[T1,T2,NODE,S](Grid2Duc[T1,T2,NODE,S]):
        Grid2Ducfs(vector[S]&, vector[triangleElem[T2]]&, T1, int,
                   vector[S]& ref_pts, int order, size_t, bool) except +

cdef extern from "Grid2Dunsp.h" namespace "ttcr" nogil:
    cdef cppclass Grid2Dunsp[T1,T2,NODE,S](Grid2Dun[T1,T2,NODE,S]):
        Grid2Dunsp(vector[S]&, vector[triangleElem[T2]]&, T2, size_t) except +

cdef extern from "Grid2Dunfs.h" namespace "ttcr" nogil:
    cdef cppclass Grid2Dunfs[T1,T2,NODE,S](Grid2Dun[T1,T2,NODE,S]):
        Grid2Dunfs(vector[S]&, vector[triangleElem[T2]]&, T1, int,
                   vector[S]& ref_pts, int order, size_t, bool) except +





cdef extern from "ttcr_t.h" namespace "ttcr" nogil:
    cdef cppclass sxyz[T]:
        sxyz(T, T, T) except +
        T x
        T y
        T z

    cdef cppclass sxz[T]:
        sxz(T, T) except +
        T x
        T z

    cdef cppclass siv[T]:
        siv(size_t, T) except +
        size_t i
        T v

    cdef cppclass siv2[T]:
        siv2(size_t, T) except +
        size_t i
        T v
        T v2

    cdef cppclass sijv[T]:
        sijv(size_t, T) except +
        size_t i
        size_t j
        T v

cdef extern from "Node3Dn.h" namespace "ttcr" nogil:
    cdef cppclass Node3Dn[T1,T2]:
        pass

cdef extern from "Node3Dnsp.h" namespace "ttcr" nogil:
    cdef cppclass Node3Dnsp[T1,T2]:
        pass

cdef extern from "Node3Dc.h" namespace "ttcr" nogil:
    cdef cppclass Node3Dc[T1,T2]:
        pass

cdef extern from "Node3Dcsp.h" namespace "ttcr" nogil:
    cdef cppclass Node3Dcsp[T1,T2]:
        pass

cdef extern from "Cell.h" namespace "ttcr" nogil:
    cdef cppclass Cell[T,NODE,S]:
        pass

cdef extern from "Node2Dn.h" namespace "ttcr" nogil:
    cdef cppclass Node2Dn[T1,T2]:
        pass

cdef extern from "Node2Dc.h" namespace "ttcr" nogil:
    cdef cppclass Node2Dc[T1,T2]:
        pass

cdef extern from "Node2Dnsp.h" namespace "ttcr" nogil:
    cdef cppclass Node2Dnsp[T1,T2]:
        pass

cdef extern from "Node2Dcsp.h" namespace "ttcr" nogil:
    cdef cppclass Node2Dcsp[T1,T2]:
        pass



# -*- coding: utf-8 -*-
"""Tests for verifying python wrappers, module tmesh in 3D"""

import unittest

import numpy as np
import vtk
from vtk.util.numpy_support import vtk_to_numpy

import ttcrpy.tmesh as tm


def get_tt(filename):
    reader = vtk.vtkXMLUnstructuredGridReader()
    reader.SetFileName(filename)
    reader.Update()
    data = reader.GetOutput()
    names = ('Travel Time', 'Travel time', 'travel time')
    for name in names:
        if data.GetPointData().HasArray(name):
            break
    tt = vtk_to_numpy(data.GetPointData().GetArray(name))
    return tt


class TestMesh3Dc(unittest.TestCase):

    def setUp(self):
        reader = vtk.vtkXMLUnstructuredGridReader()
        reader.SetFileName('layers_coarse.vtu')
        reader.Update()

        self.nodes = np.empty((reader.GetOutput().GetNumberOfPoints(), 3 ))
        for n in range(reader.GetOutput().GetNumberOfPoints()):
            x = reader.GetOutput().GetPoint(n)
            self.nodes[n, 0] = x[0]
            self.nodes[n, 1] = x[1]
            self.nodes[n, 2] = x[2]

        self.tet = np.empty((reader.GetOutput().GetNumberOfCells(), 4 ), dtype=int)
        ind = vtk.vtkIdList()
        for n in range(reader.GetOutput().GetNumberOfCells()):
            reader.GetOutput().GetCellPoints(n, ind)
            self.tet[n, 0] = ind.GetId(0)
            self.tet[n, 1] = ind.GetId(1)
            self.tet[n, 2] = ind.GetId(2)
            self.tet[n, 3] = ind.GetId(3)

        data = reader.GetOutput()
        self.slowness = vtk_to_numpy(data.GetCellData().GetArray('Slowness'))

        self.src = np.loadtxt('src.dat',skiprows=1)
        self.src = self.src.reshape((1, 4))
        self.rcv = np.loadtxt('rcv.dat',skiprows=1)

    def test_Mesh3Dfs(self):
        g = tm.Mesh3d(self.nodes, self.tet, method='FSM', tt_from_rp=0)
        tt = g.raytrace(self.src, self.rcv, slowness=self.slowness)
        tt = g.get_grid_traveltimes()
        tt = tt.flatten()
        tt_ref = get_tt('fsm_d_p_lc_u_src_all_tt.vtu')
        self.assertLess(np.sum(np.abs(tt-tt_ref))/tt.size, 0.01,
                        'FSM accuracy failed (slowness in cells)')

    def test_Mesh3Dsp(self):
        g = tm.Mesh3d(self.nodes, self.tet, method='SPM', n_secondary=5, tt_from_rp=0)
        tt = g.raytrace(self.src, self.rcv, slowness=self.slowness)
        tt = g.get_grid_traveltimes()
        tt = tt.flatten()
        tt_ref = get_tt('spm_d_p_lc_05_u_src_all_tt.vtu')
        self.assertLess(np.sum(np.abs(tt-tt_ref))/tt.size, 0.01,
                        'SPM accuracy failed (slowness in cells)')

    def test_Mesh3Ddsp(self):
        g = tm.Mesh3d(self.nodes, self.tet, method='DSPM', n_secondary=2,
                      n_tertiary=2, radius_tertiary=2.0, tt_from_rp=0)
        tt = g.raytrace(self.src, self.rcv, slowness=self.slowness)
        tt = g.get_grid_traveltimes()
        tt = tt.flatten()
        tt_ref = get_tt('dspm_d_p_lc_2_2_2_u_src_all_tt.vtu')
        self.assertLess(np.sum(np.abs(tt-tt_ref))/tt.size, 0.01,
                        'DSPM accuracy failed (slowness in cells)')


class TestMesh3Dn(unittest.TestCase):

    def setUp(self):
        reader = vtk.vtkXMLUnstructuredGridReader()
        reader.SetFileName('gradient_coarse.vtu')
        reader.Update()

        self.nodes = np.empty((reader.GetOutput().GetNumberOfPoints(), 3 ))
        for n in range(reader.GetOutput().GetNumberOfPoints()):
            x = reader.GetOutput().GetPoint(n)
            self.nodes[n, 0] = x[0]
            self.nodes[n, 1] = x[1]
            self.nodes[n, 2] = x[2]

        self.tet = np.empty((reader.GetOutput().GetNumberOfCells(), 4 ), dtype=int)
        ind = vtk.vtkIdList()
        for n in range(reader.GetOutput().GetNumberOfCells()):
            reader.GetOutput().GetCellPoints(n, ind)
            self.tet[n, 0] = ind.GetId(0)
            self.tet[n, 1] = ind.GetId(1)
            self.tet[n, 2] = ind.GetId(2)
            self.tet[n, 3] = ind.GetId(3)

        data = reader.GetOutput()
        self.slowness = vtk_to_numpy(data.GetPointData().GetArray('Slowness'))

        self.src = np.loadtxt('src.dat',skiprows=1)
        self.src = self.src.reshape((1, 4))
        self.rcv = np.loadtxt('rcv.dat',skiprows=1)

    def test_Mesh3Dfs(self):
        g = tm.Mesh3d(self.nodes, self.tet, method='FSM', cell_slowness=0, tt_from_rp=0)
        tt = g.raytrace(self.src, self.rcv, slowness=self.slowness)
        tt = g.get_grid_traveltimes()
        tt = tt.flatten()
        tt_ref = get_tt('fsm_d_p_gc_u_src_all_tt.vtu')
        self.assertLess(np.sum(np.abs(tt-tt_ref))/tt.size, 0.01,
                        'FSM accuracy failed (slowness at nodes)')

    def test_Mesh3Dsp(self):
        g = tm.Mesh3d(self.nodes, self.tet, method='SPM', n_secondary=5, cell_slowness=0, tt_from_rp=0)
        tt = g.raytrace(self.src, self.rcv, slowness=self.slowness)
        tt = g.get_grid_traveltimes()
        tt = tt.flatten()
        tt_ref = get_tt('spm_d_p_gc_05_u_src_all_tt.vtu')
        self.assertLess(np.sum(np.abs(tt-tt_ref))/tt.size, 0.01,
                        'SPM accuracy failed (slowness at nodes)')

    def test_Mesh3Ddsp(self):
        g = tm.Mesh3d(self.nodes, self.tet, method='DSPM', n_secondary=2,
                      n_tertiary=2, radius_tertiary=2.0, cell_slowness=0, tt_from_rp=0)
        tt = g.raytrace(self.src, self.rcv, slowness=self.slowness)
        tt = g.get_grid_traveltimes()
        tt = tt.flatten()
        tt_ref = get_tt('dspm_d_p_gc_2_2_2_u_src_all_tt.vtu')
        self.assertLess(np.sum(np.abs(tt-tt_ref))/tt.size, 0.01,
                        'DSPM accuracy failed (slowness at nodes)')

#
# class Data_kernel(unittest.TestCase):
#
#     def test_2d(self):
#
#         V = np.ones((11, 13))
#         V[:, 7:] = 2
#         slowness = 1. / V.flatten()
#
#         grx = np.arange(12.)
#         grz = np.arange(14.)
#
#         z = 0.5 + np.arange(13.)
#         Tx = np.vstack((0.5+np.zeros((13,)),
#                         z)).T
#         Rx = np.vstack((10.5+np.zeros((13,)),
#                         z)).T
#         nTx = Tx.shape[0]
#         nRx = Rx.shape[0]
#         Tx = np.kron(Tx, np.ones((nRx,1)))
#         Rx = np.kron(np.ones((nTx,1)), Rx)
#
#         L = tm.Mesh2d.data_kernel_straight_rays(Tx, Rx, grx, grz)
#         tt = L.dot(slowness)
#
#         tt2 = np.zeros(tt.shape)
#         d = np.sqrt(np.sum((Tx-Rx)**2, axis=1))
#
#         ind = np.logical_and(Tx[:,1]>7, Rx[:,1]>7)
#         tt2[ind] = d[ind]/2
#
#         ind2 = np.logical_and(Tx[:,1]<7, Rx[:,1]<7)
#         tt2[ind2] = d[ind2]
#
#         ind3 = np.logical_and(np.logical_not(ind), np.logical_not(ind2))
#
#         f = (7-Tx[ind3,1]) / (Rx[ind3,1]-Tx[ind3,1])
#         ind = (Rx[ind3,1]-Tx[ind3,1]) < 0
#         f[ind] = 1-f[ind]
#         tt2[ind3] = d[ind3]*f + d[ind3]*(1-f)/2
#
#         self.assertAlmostEqual(np.sum(np.abs(tt-tt2)), 0.0 )

if __name__ == '__main__':

    unittest.main()



# -*- coding: utf-8 -*-
"""
Raytracing on unstructured triangular and tetrahedral meshes

This module contains two classes to perform traveltime computation and
raytracing on unstructured meshes:
    - `Mesh2d` for 2D media
    - `Mesh3d` for 3D media

    Two general algorithms are implemented
        - the Shortest-Path Method
        - the Fast-Sweeping Method
    and `Mesh3d` also support the Dynamic Shortest-Path Method

    Slowness model can be defined in two ways:
        1) slowness constant within the voxels of the mesh (the default)
        2) slowness defined at nodes of the mesh

    This code is part of ttcr ( https://github.com/groupeLIAMG/ttcr )
"""

# distutils: language = c++

import numpy as np
cimport numpy as np
import scipy.sparse as sp

import vtk
from vtk.util import numpy_support

from ttcrpy.tmesh cimport Grid3D, Grid3Ducfs, Grid3Ducsp, Grid3Ducdsp, \
    Grid3Dunfs, Grid3Dunsp, Grid3Dundsp, Grid2D, Grid2Duc, Grid2Dun, \
    Grid2Ducsp, Grid2Ducfs, Grid2Dunsp, Grid2Dunfs

cdef extern from "verbose.h" namespace "ttcr" nogil:
    void setVerbose(int)

cdef extern from "utils_cython.h":
    int build_matrix_siv[T](size_t, size_t, vector[vector[siv[T]]]&, object)

def set_verbose(v):
    """Set verbosity level for C++ code

    Parameters
    ----------
    v: int
        verbosity level
    """
    setVerbose(v)


cdef class Mesh3d:
    """class to perform raytracing with tetrahedral meshes

    Attributes
    ----------
    nparams: int
        total number of parameters for grid
    n_threads: int
        number of threads for raytracing

    Constructor:

    Mesh3d(nodes, tetra, n_threads, cell_slowness, method, gradient_method, tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary, n_tertiary, radius_tertiary) -> Mesh3d

        Parameters
        ----------
        nodes : np.ndarray, shape (nnodes, 3)
            node coordinates
        tetra : np.ndarray of int, shape (ntetra, 4)
            indices of nodes forming the tetrahedra
        n_threads : int
            number of threads for raytracing (default is 1)
        cell_slowness : bool
            slowness defined for cells (True) or nodes (False) (default is 1)
        method : string
            raytracing method (default is FSM)
                - 'FSM' : fast marching method
                - 'SPM' : shortest path method
                - 'DSPM' : dynamic shortest path
        gradient_method : int
            method to compute traveltime gradient (default is 1)
                - 0 : least-squares first-order
                - 1 : least-squares second-order
                - 2 : Averaging-Based method
        tt_from_rp : bool
            compute traveltimes from raypaths (SPM od DSPM) (default is 1)
        interp_vel : bool
            interpolate velocity instead of slowness at nodes (for
            cell_slowness == False or FSM) (defauls is False)
        eps : double
            convergence criterion (FSM) (default is 1e-15)
        maxit : int
            max number of sweeping iterations (FSM) (default is 20)
        min_dist : double
            tolerance for backward raytracing (default is 1e-5)
        n_secondary : int
            number of secondary nodes (SPM & DSPM) (default is 2)
        n_tertiary : int
            number of tertiary nodes (DSPM) (default is 2)
        radius_tertiary : double
            radius of sphere around source that includes tertiary nodes (DSPM)
            (default is 1)

    """
    cdef bool cell_slowness
    cdef size_t _n_threads
    cdef char method
    cdef bool tt_from_rp
    cdef bool interp_vel
    cdef double eps
    cdef int maxit
    cdef int gradient_method
    cdef double min_dist
    cdef uint32_t n_secondary
    cdef uint32_t n_tertiary
    cdef double radius_tertiary
    cdef vector[sxyz[double]] no
    cdef vector[tetrahedronElem[uint32_t]] tet
    cdef Grid3D[double, uint32_t]* grid

    def __cinit__(self, np.ndarray[np.double_t, ndim=2] nodes,
                  np.ndarray[np.int64_t, ndim=2] tetra,
                  size_t n_threads=1, bool cell_slowness=1,
                  str method='FSM', int gradient_method=1,
                  bool tt_from_rp=1, bool interp_vel=0,
                  double eps=1.e-15, int maxit=20, double min_dist=1.e-5,
                  uint32_t n_secondary=2, uint32_t n_tertiary=2,
                  double radius_tertiary=1.0):

        self.cell_slowness = cell_slowness
        self._n_threads = n_threads
        self.tt_from_rp = tt_from_rp
        self.interp_vel = interp_vel
        self.eps = eps
        self.maxit = maxit
        self.gradient_method = gradient_method
        self.min_dist = min_dist
        self.n_secondary = n_secondary
        self.n_tertiary = n_tertiary
        self.radius_tertiary = radius_tertiary

        cdef double source_radius = 0.0
        cdef vector[sxyz[double]] pts_ref
        cdef int n

        if not nodes.flags['C_CONTIGUOUS']:
            nodes = np.ascontiguousarray(nodes)
        if not tetra.flags['C_CONTIGUOUS']:
            tetra = np.ascontiguousarray(tetra)

        for n in range(nodes.shape[0]):
            self.no.push_back(sxyz[double](nodes[n, 0],
                                           nodes[n, 1],
                                           nodes[n, 2]))
        for n in range(tetra.shape[0]):
            self.tet.push_back(tetrahedronElem[uint32_t](tetra[n, 0],
                                                         tetra[n, 1],
                                                         tetra[n, 2],
                                                         tetra[n, 3]))

        if method == 'FSM':
            xmin = np.min(nodes[:, 0])
            xmax = np.max(nodes[:, 0])
            ymin = np.min(nodes[:, 1])
            ymax = np.max(nodes[:, 1])
            zmin = np.min(nodes[:, 2])
            zmax = np.max(nodes[:, 2])
            pts_ref.push_back(sxyz[double](xmin, ymin, zmin))
            pts_ref.push_back(sxyz[double](xmin, ymin, zmax))
            pts_ref.push_back(sxyz[double](xmin, ymax, zmin))
            pts_ref.push_back(sxyz[double](xmin, ymax, zmax))
            pts_ref.push_back(sxyz[double](xmax, ymin, zmin))
            pts_ref.push_back(sxyz[double](xmax, ymin, zmax))
            pts_ref.push_back(sxyz[double](xmax, ymax, zmin))
            pts_ref.push_back(sxyz[double](xmax, ymax, zmax))

        if cell_slowness:
            if method == 'FSM':
                self.method = b'f'
                self.grid = new Grid3Ducfs[double,uint32_t](self.no, self.tet,
                                                            eps, maxit,
                                                            pts_ref, 2,
                                                            gradient_method,
                                                            tt_from_rp,
                                                            min_dist, n_threads)
            elif method == 'SPM':
                self.method = b's'
                self.grid = new Grid3Ducsp[double,uint32_t](self.no, self.tet,
                                                            n_secondary,
                                                            tt_from_rp,
                                                            min_dist, n_threads)
            elif method == 'DSPM':
                self.method = b'd'
                self.grid = new Grid3Ducdsp[double,uint32_t](self.no, self.tet,
                                                             n_secondary,
                                                             n_tertiary,
                                                             source_radius,
                                                             gradient_method,
                                                             tt_from_rp,
                                                             min_dist,
                                                             radius_tertiary,
                                                             n_threads)

            else:
                raise ValueError('Method {0:s} undefined'.format(method))
        else:
            if method == 'FSM':
                self.method = b'f'
                self.grid = new Grid3Dunfs[double,uint32_t](self.no, self.tet,
                                                            eps, maxit,
                                                            pts_ref, 2,
                                                            gradient_method,
                                                            interp_vel,
                                                            tt_from_rp,
                                                            min_dist, n_threads)
            elif method == 'SPM':
                self.method = b's'
                self.grid = new Grid3Dunsp[double,uint32_t](self.no, self.tet,
                                                            n_secondary,
                                                            interp_vel,
                                                            tt_from_rp,
                                                            min_dist, n_threads)
            elif method == 'DSPM':
                self.method = b'd'
                self.grid = new Grid3Dundsp[double,uint32_t](self.no, self.tet,
                                                             n_secondary,
                                                             n_tertiary,
                                                             source_radius,
                                                             interp_vel,
                                                             gradient_method,
                                                             tt_from_rp,
                                                             min_dist,
                                                             radius_tertiary,
                                                             n_threads)
            else:
                raise ValueError('Method {0:s} undefined'.format(method))

    def __dealloc__(self):
        del self.grid

    def __reduce__(self):
        if self.method == b'f':
            method = 'FSM'
        elif self.method == b's':
            method = 'SPM'
        elif self.method == b'd':
            method = 'DSPM'

        nodes = np.ndarray((self.no.size(), 3))
        tetra = np.ndarray((self.tet.size(), 4), dtype=int)
        cdef int n
        cdef int nn
        for n in range(nodes.shape[0]):
            nodes[n, 0] = self.no[n].x
            nodes[n, 1] = self.no[n].y
            nodes[n, 2] = self.no[n].z
        for n in range(tetra.shape[0]):
            for nn in range(4):
                tetra[n, nn] = self.tet[n].i[nn]

        constructor_params = (nodes, tetra, method, self.cell_slowness,
                              self._n_threads, self.tt_from_rp, self.interp_vel,
                              self.eps, self.maxit, self.gradient_method,
                              self.min_dist, self.n_secondary, self.n_tertiary,
                              self.radius_tertiary)
        return (_rebuild3d, constructor_params)

    @property
    def n_threads(self):
        """int: number of threads for raytracing"""
        return self._n_threads

    @property
    def nparams(self):
        """int: total number of parameters for mesh"""
        if self.cell_slowness:
            return self.tet.size()
        else:
            return self.no.size()

    def set_use_thread_pool(self, use_thread_pool):
        """bool: use thread pool instead of parallel loop"""
        self.grid.setUsePool(use_thread_pool)

    def is_outside(self, np.ndarray[np.double_t, ndim=2] pts):
        """
        is_outside(pts)

        Check if points are outside grid

        Parameters
        ----------
        pts : np ndarray, shape(npts, 3)
            coordinates of points to check

        Returns
        -------
        bool:
            True if at least one point outside grid
        """
        cdef vector[sxyz[double]] vpts
        for n in range(pts.shape[0]):
            vpts.push_back(sxyz[double](pts[n, 0], pts[n, 1], pts[n, 2]))
        try:
            self.grid.checkPts(vpts)  # throws an except if 1 pt is outside
        except RuntimeError:
            return True

        return False

    def get_number_of_nodes(self):
        """
        Returns
        -------
        int:
            number of nodes in grid
        """
        return self.no.size()

    def get_number_of_cells(self):
        """
        Returns
        -------
        int:
            number of cells in grid
        """
        return self.tet.size()

    def get_grid_traveltimes(self, thread_no=0):
        """
        get_grid_traveltimes(thread_no=0)

        Obtain traveltimes computed at primary grid nodes

        Parameters
        ----------
        thread_no : int
            thread used to computed traveltimes (default is 0)

        Returns
        -------
        tt: np ndarray, shape (nnodes,)
            traveltimes
        """
        if thread_no >= self._n_threads:
            raise ValueError('Thread number is larger than number of threads')
        cdef vector[double] tmp
        cdef int n
        self.grid.getTT(tmp, thread_no)
        tt = np.empty((tmp.size(),))
        for n in range(tmp.size()):
            tt[n] = tmp[n]
        return tt

    def set_slowness(self, slowness):
        """
        set_slowness(slowness)

        Assign slowness to grid

        Parameters
        ----------
        slowness : np ndarray, shape (nparams, )
        """
        if slowness.size != self.nparams:
            raise ValueError('Slowness vector has wrong size')

        if not slowness.flags['C_CONTIGUOUS']:
            slowness = np.ascontiguousarray(slowness)

        cdef vector[double] slown
        cdef int i
        for i in range(slowness.size):
            slown.push_back(slowness[i])
        self.grid.setSlowness(slown)

    def set_velocity(self, velocity):
        """
        set_velocity(velocity)

        Assign velocity to grid

        Parameters
        ----------
        velocity : np ndarray, shape (nparams, )
        """
        if velocity.size != self.nparams:
            raise ValueError('velocity vector has wrong size')

        if not velocity.flags['C_CONTIGUOUS']:
            velocity = np.ascontiguousarray(velocity)

        cdef vector[double] slown
        cdef int i
        for i in range(velocity.size):
            slown.push_back(1./velocity[i])
        self.grid.setSlowness(slown)

    def compute_D(self, coord):
        """
        compute_D(coord)

        Return matrix of interpolation weights for velocity data points
        constraint

        Parameters
        ----------
        coord : np.ndarray, shape (npts, 3)
            coordinates of data points

        Returns
        -------
        D : scipy csr_matrix, shape (npts, nparams)
            Matrix of interpolation weights
        """
        if self.is_outside(coord):
            raise ValueError('Velocity data point outside grid')
        cdef vector[sxyz[double]] vpts
        for n in range(coord.shape[0]):
            vpts.push_back(sxyz[double](coord[n, 0], coord[n, 1], coord[n, 2]))

        cdef vector[vector[sijv[double]]] d_data

        self.grid.computeD(vpts, d_data)

        nnz = 0
        for ni in range(d_data.size()):
            nnz += d_data[ni].size()

        MM = vpts.size()
        NN = self.nparams
        indptr = np.empty((MM+1,), dtype=np.int64)
        indices = np.empty((nnz,), dtype=np.int64)
        val = np.empty((nnz,))

        k = 0
        for i in range(MM):
            indptr[i] = k
            for j in range(NN):
                for nn in range(d_data[i].size()):
                    if d_data[i][nn].i == i and d_data[i][nn].j == j:
                        indices[k] = j
                        val[k] = d_data[i][nn].v
                        k += 1

        indptr[MM] = k
        return sp.csr_matrix((val, indices, indptr), shape=(MM,NN))

    def compute_K(self, order=2, taylor_order=2, weighting=True, squared=True,
                  s0inside=False):
        """
        Compute smoothing matrices (spatial derivative)

        Parameters
        ----------
        order : int
            order of derivative (1 or 2, 2 by default)
        taylor_order : int
            order of taylors series expansion (1 or 2, 2 by default)
        weighting : bool
            apply inverse distance weighting (True by default)
        squared : bool
            Second derivative evaluated by taking the square of first
            derivative.  Applied only if order == 2 (True by default)
        s0inside : bool
            (experimental) ignore slowness value at local node (value is a
            filtered estimate) (False by default)

        Returns
        -------
        Kx, Ky, Kz : :obj:`tuple` of :obj:`csr_matrix`
            matrices for derivatives along x, y, & z
        """

        cdef int o = order
        cdef int to = taylor_order
        cdef bool w = weighting
        cdef bool s = s0inside
        cdef vector[vector[vector[siv[double]]]] k_data

        if order == 2 and squared:
            o = 1
        self.grid.computeK(k_data, o, to, w, s)

        K = []
        cdef size_t MM = self.nparams
        cdef size_t NN = self.nparams
        for nk in range(3):
            m_tuple = ([0.0], [0.0], [0.0])
            build_matrix_siv(MM, NN, k_data[nk], m_tuple)
            K.append( sp.csr_matrix(m_tuple, shape=(MM,NN)) )
            if order == 2 and squared:
                K[-1] = K[-1] * K[-1]

        return tuple(K)

    def get_s0(self, hypo, slowness=None):
        """
        get_s0(hypo, slowness=None)

        Return slowness at source points

        Parameters
        ----------
        hypo : np.ndarray with 5 columns
            hypo holds source information, i.e.
                - 1st column is event ID number
                - 2nd column is origin time
                - 3rd column is source easting
                - 4th column is source northing
                - 5th column is source elevation
        slowness : np ndarray, shape (nparams, ) (optional)
            slowness at grid nodes or cells (depending on cell_slowness)

        Returns
        -------
        s0 : np.ndarray
            slowness at source points
        """

        cdef Py_ssize_t n, nn

        if hypo.shape[1] != 5:
            raise ValueError('hypo should be npts x 5')
        src = hypo[:,2:5]
        evID = hypo[:,0]
        eid = np.sort(np.unique(evID))
        nTx = len(eid)

        if slowness is not None:
            self.set_slowness(slowness)

        i0 = np.empty((nTx,), dtype=np.int64)
        for n in range(nTx):
            for nn in range(evID.size):
                if eid[n] == evID[nn]:
                    i0[n] = nn
                    break

        cdef vector[vector[sxyz[double]]] vTx
        vTx.resize(nTx)

        iTx = []
        i0 = 0
        for n in range(nTx):
            for nn in range(evID.size):
                if eid[n] == evID[nn]:
                    i0 = nn
                    break
            vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))

        for i in eid:
            ii = evID == i
            iTx.append(np.nonzero(ii)[0])

        s0 = np.zeros((src.shape[0],))

        for n in range(nTx):
            s = 0.0
            for nn in range(vTx[n].size()):
                s += self.grid.computeSlowness(vTx[n][nn])
            s0[iTx[n]] = s/vTx[n].size()
        return s0

    def raytrace(self, source, rcv, slowness=None, thread_no=None,
                 aggregate_src=False, compute_L=False, return_rays=False):
        """
        raytrace(source, rcv, slowness=None, thread_no=None, aggregate_src=False, compute_L=False, return_rays=False) -> tt, rays, L

        Perform raytracing

        Parameters
        ----------
        source : 2D np.ndarray with 3, 4 or 5 columns
            see notes below
        rcv : 2D np.ndarray with 3 columns
            Columns correspond to x, y and z coordinates
        slowness : np ndarray, (None by default)
            slowness at grid nodes or cells (depending on cell_slowness)
            if None, slowness must have been assigned previously
        thread_no : int (None by default)
            Perform calculations in thread number "thread_no"
            if None, attempt to run in parallel if warranted by number of
            sources and value of n_threads in constructor
        aggregate_src : bool (False by default)
            if True, all source coordinates belong to a single event
        compute_L : bool (False by default)
            Compute matrices of partial derivative of travel time w/r to slowness
        return_rays : bool (False by default)
            Return raypaths

        Returns
        -------
        tt : np.ndarray
            travel times for the appropriate source-rcv  (see Notes below)
        rays : :obj:`list` of :obj:`np.ndarray`
            Coordinates of segments forming raypaths (if return_rays is True)
        L :  :obj:`list` of :obj:`csr_matrix`  or  scipy csr_matrix
            Matrix of partial derivative of travel time w/r to slowness.
            if input argument source has 5 columns or if slowness is defined at
            nodes, L is a list of matrices and the number of matrices is equal
            to the number of sources otherwise, L is a single csr_matrix

        Notes
        -----
        If source has 3 columns:
            - Columns correspond to x, y and z coordinates
            - Origin time (t0) is 0 for all points
        If source has 4 columns:
            - 1st column corresponds to origin times
            - 2nd, 3rd & 4th columns correspond to x, y and z coordinates
        If source has 5 columns:
            - 1st column corresponds to event ID
            - 2nd column corresponds to origin times
            - 3rd, 4th & 5th columns correspond to x, y and z coordinates

        For the latter case (5 columns), source and rcv should contain the same
        number of rows, each row corresponding to a source-receiver pair.
        For the 2 other cases, source and rcv can contain the same number of
        rows, each row corresponding to a source-receiver pair, or the number
        of rows may differ if aggregate_src is True or if all rows in source
        are identical.
        """

        # check input data consistency

        if source.ndim != 2 or rcv.ndim != 2:
            raise ValueError('source and rcv should be 2D arrays')

        if self.method == b'd' and aggregate_src:
            raise ValueError('Cannot aggregate source with DSPM raytracing')

        evID = None
        if source.shape[1] == 5:
            src = source[:,2:5]
            t0 = source[:,1]
            evID = source[:,0]
            eid = np.sort(np.unique(evID))
            nTx = len(eid)
        elif source.shape[1] == 3:
            src = source
            Tx = np.unique(source, axis=0)
            t0 = np.zeros((Tx.shape[0], 1))
            nTx = Tx.shape[0]
        elif source.shape[1] == 4:
            src = source[:,1:4]
            tmp = np.unique(source, axis=0)
            nTx = tmp.shape[0]
            Tx = tmp[:,1:4]
            t0 = tmp[:,0]
        else:
            raise ValueError('source should be either nsrc x 3, 4 or 5')

        if src.shape[1] != 3 or rcv.shape[1] != 3:
            raise ValueError('src and rcv should be ndata x 3')

        if self.is_outside(src):
            raise ValueError('Source point outside grid')

        if self.is_outside(rcv):
            raise ValueError('Receiver outside grid')

        if slowness is not None:
            self.set_slowness(slowness)

        cdef vector[vector[sxyz[double]]] vTx
        cdef vector[vector[sxyz[double]]] vRx
        cdef vector[vector[double]] vt0
        cdef vector[vector[double]] vtt

        cdef vector[vector[vector[sxyz[double]]]] r_data
        cdef vector[vector[vector[siv[double]]]] l_data
        cdef vector[vector[vector[sijv[double]]]] m_data
        cdef size_t thread_nb

        cdef int i, n, n2, nt

        vTx.resize(nTx)
        vRx.resize(nTx)
        vt0.resize(nTx)
        vtt.resize(nTx)
        if compute_L and self.cell_slowness:
            raise NotImplementedError('compute_L not implemented for mesh with slowness defined in cells')
        elif compute_L and not self.cell_slowness:
            m_data.resize(nTx)

        if return_rays:
            r_data.resize(nTx)

        iRx = []
        if evID is None:
            if nTx == 1:
                vTx[0].push_back(sxyz[double](src[0,0], src[0,1], src[0,2]))
                for r in rcv:
                    vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
                vt0[0].push_back(t0[0])
                vtt[0].resize(rcv.shape[0])
                iRx.append(np.arange(rcv.shape[0]))
            elif aggregate_src:
                for t in Tx:
                    vTx[0].push_back(sxyz[double](t[0], t[1], t[2]))
                for t in t0:
                    vt0[0].push_back(t)
                for r in rcv:
                    vRx[0].push_back(sxyz[double](r[0], r[1], r[2]))
                vtt[0].resize(rcv.shape[0])
                nTx = 1
                iRx.append(np.arange(rcv.shape[0]))
            else:
                if src.shape != rcv.shape:
                    raise ValueError('src and rcv should be of equal size')

                for n in range(nTx):
                    ind = np.sum(Tx[n,:] == src, axis=1) == 3
                    iRx.append(np.nonzero(ind)[0])
                    vTx[n].push_back(sxyz[double](Tx[n,0], Tx[n,1], Tx[n,2]))
                    vt0[n].push_back(t0[n])
                    for r in rcv[ind,:]:
                        vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
                    vtt[n].resize(vRx[n].size())
        else:
            if src.shape != rcv.shape:
                raise ValueError('src and rcv should be of equal size')

            i0 = 0
            for n in range(nTx):
                for nn in range(evID.size):
                    if eid[n] == evID[nn]:
                        i0 = nn
                        break
                vTx[n].push_back(sxyz[double](src[i0,0], src[i0,1], src[i0,2]))
                vt0[n].push_back(t0[i0])

            for i in eid:
                ii = evID == i
                iRx.append(np.nonzero(ii)[0])

            for n in range(nTx):
                for r in rcv[iRx[n],:]:
                    vRx[n].push_back(sxyz[double](r[0], r[1], r[2]))
                vtt[n].resize(vRx[n].size())

        tt = np.zeros((rcv.shape[0],))
        if nTx < self._n_threads or self._n_threads == 1:
            if compute_L==False and return_rays==False:
                for n in range(nTx):
                    self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
            elif compute_L and return_rays:
                if self.cell_slowness:
                    # TODO: implement this in C++ codebase!
                    for n in range(nTx):
                        self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], l_data[n], 0)
                else:
                    for n in range(nTx):
                        self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], m_data[n], 0)
            elif compute_L:
                if self.cell_slowness:
                    for n in range(nTx):
                        self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], l_data[n], 0)
                else:
                    for n in range(nTx):
                        self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], m_data[n], 0)
            else:
                for n in range(nTx):
                    self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)

        elif thread_no is not None:
            # we should be here for just one event
            assert nTx == 1
            # normally we should not need to compute M or L
            assert compute_L is False
            thread_nb = thread_no

            if return_rays:
                self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
                for nt in range(vtt[0].size()):
                    tt[nt] = vtt[0][nt]
                rays = []
                for n2 in range(vRx.size()):
                    r = np.empty((r_data[0][n2].size(), 3))
                    for nn in range(r_data[0][n2].size()):
                        r[nn, 0] = r_data[0][n2][nn].x
                        r[nn, 1] = r_data[0][n2][nn].y
                        r[nn, 2] = r_data[0][n2][nn].z
                    rays.append(r)
                return tt, rays

            else:
                self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)
                for nt in range(vtt[0].size()):
                    tt[nt] = vtt[0][nt]
                return tt

        else:
            if compute_L==False and return_rays==False:
                self.grid.raytrace(vTx, vt0, vRx, vtt)
            elif compute_L and return_rays:
                if self.cell_slowness:
                    self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, l_data)
                else:
                    self.grid.raytrace(vTx, vt0, vRx, vtt, r_data, m_data)
            elif compute_L:
                if self.cell_slowness:
                    self.grid.raytrace(vTx, vt0, vRx, vtt, l_data)
                else:
                    self.grid.raytrace(vTx, vt0, vRx, vtt, m_data)
            else:
                self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)

        for n in range(nTx):
            for nt in range(vtt[n].size()):
                tt[iRx[n][nt]] = vtt[n][nt]

        if return_rays:
            rays = [ [0.0] for n in range(rcv.shape[0])]
            for n in range(nTx):
                r = [ [0.0] for i in range(vRx[n].size())]
                for n2 in range(vRx[n].size()):
                    r[n2] = np.empty((r_data[n][n2].size(), 3))
                    for nn in range(r_data[n][n2].size()):
                        r[n2][nn, 0] = r_data[n][n2][nn].x
                        r[n2][nn, 1] = r_data[n][n2][nn].y
                        r[n2][nn, 2] = r_data[n][n2][nn].z
                for nt in range(vtt[n].size()):
                    rays[iRx[n][nt]] = r[nt]

        if compute_L and not self.cell_slowness:
            # we return array of matrices, one for each event
            L = []
            for n in range(nTx):
                nnz = 0
                for ni in range(m_data[n].size()):
                    nnz += m_data[n][ni].size()
                indptr = np.empty((vRx[n].size()+1,), dtype=np.int64)
                indices = np.empty((nnz,), dtype=np.int64)
                val = np.empty((nnz,))

                k = 0
                MM = vRx[n].size()
                NN = self.get_number_of_nodes()
                for i in range(MM):
                    indptr[i] = k
                    for j in range(NN):
                        for nn in range(m_data[n][i].size()):
                            if m_data[n][i][nn].i == i and m_data[n][i][nn].j == j:
                                indices[k] = j
                                val[k] = m_data[n][i][nn].v
                                k += 1

                indptr[MM] = k
                L.append( sp.csr_matrix((val, indices, indptr), shape=(MM,NN)) )

        if compute_L==False and return_rays==False:
            return tt
        elif compute_L and return_rays:
            return tt, rays, L
        elif compute_L:
            return tt, L
        else:
            return tt, rays

    def to_vtk(self, fields, filename):
        """
        to_vtk(fields, filename)

        Save mesh variables and/or raypaths to VTK format

        Parameters
        ----------
        fields: dict
            dict of variables to save to file. Variables should be np.ndarray of
            size equal to either the number of nodes of the number of cells of
            the mesh, or a list of raypath coordinates.
        filename: str
            Name of file without extension for saving (extension vtu will be
            added).  Raypaths are saved in separate files, and filename will
            be appended by the dict key and have a vtp extension.

        Notes
        -----
        VTK files can be visualized with Paraview (https://www.paraview.org)
        """
        cdef int n, nn
        ugrid = vtk.vtkUnstructuredGrid()
        tPts = vtk.vtkPoints()
        tPts.SetNumberOfPoints(self.no.size())
        for n in range(self.no.size()):
            tPts.InsertPoint(n, self.no[n].x, self.no[n].y, self.no[n].z)
        ugrid.SetPoints(tPts)
        tet = vtk.vtkTetra()
        for n in range(self.tet.size()):
            for nn in range(4):
                tet.GetPointIds().SetId(nn, self.tet[n].i[nn])
            ugrid.InsertNextCell(tet.GetCellType(), tet.GetPointIds())

        save_grid = False
        for fn in fields:
            data = fields[fn]

            if isinstance(data, list):
                self._save_raypaths(data, filename+'_'+fn+'.vtp')
            else:
                save_grid = True
                scalar = vtk.vtkDoubleArray()
                scalar.SetName(fn)
                scalar.SetNumberOfComponents(1)
                scalar.SetNumberOfTuples(data.size)
                if data.size == self.get_number_of_nodes():
                    for n in range(data.size):
                        scalar.SetTuple1(n, data[n])
                    ugrid.GetPointData().AddArray(scalar)
                elif data.size == self.get_number_of_cells():
                    for n in range(data.size):
                        scalar.SetTuple1(n, data[n])
                    ugrid.GetCellData().AddArray(scalar)
                else:
                    raise ValueError('Field {0:s} has incorrect size'.format(fn))

        if save_grid:
            writer = vtk.vtkXMLUnstructuredGridWriter()
            writer.SetFileName(filename+'.vtu')
            writer.SetInputData(ugrid)
            writer.SetDataModeToBinary()
            writer.Update()

    def  _save_raypaths(self, rays, filename):
        polydata = vtk.vtkPolyData()
        cellarray = vtk.vtkCellArray()
        pts = vtk.vtkPoints()
        npts = 0
        for n in range(len(rays)):
            npts += rays[n].shape[0]
        pts.SetNumberOfPoints(npts)
        npts = 0
        for n in range(len(rays)):
            for p in range(rays[n].shape[0]):
                pts.InsertPoint(npts, rays[n][p, 0], rays[n][p, 1], rays[n][p, 2])
                npts += 1
        polydata.SetPoints(pts)
        npts = 0
        for n in range(len(rays)):
            line = vtk.vtkPolyLine()
            line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
            for p in range(rays[n].shape[0]):
                line.GetPointIds().SetId(p, npts)
                npts += 1
            cellarray.InsertNextCell(line)
        polydata.SetLines(cellarray)
        writer = vtk.vtkXMLPolyDataWriter()
        writer.SetFileName(filename)
        writer.SetInputData(polydata)
        writer.SetDataModeToBinary()
        writer.Update()


    @staticmethod
    def builder(filename, size_t n_threads=1, bool cell_slowness=1,
                str method='FSM', int gradient_method=1,
                bool tt_from_rp=1, bool interp_vel=0,
                double eps=1.e-15, int maxit=20, double min_dist=1.e-5,
                uint32_t n_secondary=2, uint32_t n_tertiary=2,
                double radius_tertiary=1.0):
        """
        builder(filename, n_threads, cell_slowness, method, gradient_method, tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary, n_tertiary, radius_tertiary)

        Build instance of Mesh3d from VTK file

        Parameters
        ----------
        filename : str
            Name of file holding a vtkUnstructuredGrid.
            The grid must have point or cell attribute named either
            'Slowness', 'slowness', 'Velocity', 'velocity', or
            'P-wave velocity'.  All cells must be of type vtkTetra

        Other parameters are defined in Constructor

        Returns
        -------
        mesh: :obj:`Mesh3d`
            mesh instance
        """

        cdef int n, nn
        reader = vtk.vtkXMLUnstructuredGridReader()
        reader.SetFileName(filename)
        reader.Update()

        data = reader.GetOutput()
        nodes = numpy_support.vtk_to_numpy(data.GetPoints().GetData())
        for n in range(data.GetNumberOfCells()):
            if data.GetCellType(n) != vtk.VTK_TETRA:
                raise ValueError('{0:s} should only contain tetrahedra')
        tet = np.ndarray((data.GetNumberOfCells(), 4))
        for n in range(data.GetNumberOfCells()):
            for nn in range(4):
                tet[n, nn] = data.GetCell(n).GetPointIds().GetId(nn)

        names = ('Slowness', 'slowness', 'Velocity', 'velocity',
                 'P-wave velocity')
        for name in names:
            if data.GetPointData().HasArray(name):
                cell_slowness = 0
                data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
                break
            if data.GetCellData().HasArray(name):
                cell_slowness = 1
                data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
                break
        else:
            raise ValueError('File should contain slowness or velocity data')

        if 'lowness' in name:
            slowness = data
        else:
            slowness = 1.0 / data

        m = Mesh3d(nodes, tet, n_threads, cell_slowness, method, gradient_method,
                   tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary,
                   n_tertiary, radius_tertiary)
        m.set_slowness(slowness)
        return m


cdef class Mesh2d:
    """class to perform raytracing with triangular meshes

    Attributes
    ----------
    nparams: int
        total number of parameters for grid
    n_threads: int
        number of threads for raytracing

    Constructor:

    Mesh2d(nodes, triangles, n_threads, cell_slowness, method, eps, maxit, process_obtuse, n_secondary) -> Mesh2d

        Parameters
        ----------
        nodes : np.ndarray, shape (nnodes, 2)
            node coordinates
        triangles : np.ndarray of int, shape (ntriangles, 3)
            indices of nodes forming the triangles
        n_threads : int
            number of threads for raytracing (default is 1)
        cell_slowness : bool
            slowness defined for cells (True) or nodes (False) (default is 1)
        method : string
            raytracing method (default is FSM)
                - 'FSM' : fast marching method
                - 'SPM' : shortest path method
                - 'DSPM' : dynamic shortest path
        eps : double
            convergence criterion (FSM) (default is 1e-15)
        maxit : int
            max number of sweeping iterations (FSM) (default is 20)
        process_obtuse : bool
            use method of Qian et al (2007) to improve accuracy for triangles
            with obtuse angle
        n_secondary : int
            number of secondary nodes (SPM) (default is 5)
    """
    cdef bool cell_slowness
    cdef bool process_obtuse
    cdef size_t _n_threads
    cdef double eps
    cdef int maxit
    cdef char method
    cdef uint32_t n_secondary
    cdef vector[sxz[double]] no
    cdef vector[triangleElem[uint32_t]] tri
    cdef Grid2D[double, uint32_t,sxz[double]]* grid

    def __cinit__(self, np.ndarray[np.double_t, ndim=2] nodes,
                  np.ndarray[np.int64_t, ndim=2] triangles,
                  size_t n_threads=1, bool cell_slowness=1,
                  str method='FSM', double eps=1.e-15, int maxit=20,
                  bool process_obtuse=1, uint32_t n_secondary=5):

        self.cell_slowness = cell_slowness
        self._n_threads = n_threads
        self.eps = eps
        self.maxit = maxit
        self.process_obtuse = process_obtuse
        self.n_secondary = n_secondary

        cdef vector[sxz[double]] pts_ref
        cdef int n

        if not nodes.flags['C_CONTIGUOUS']:
            nodes = np.ascontiguousarray(nodes)
        if not triangles.flags['C_CONTIGUOUS']:
            triangles = np.ascontiguousarray(triangles)

        for n in range(nodes.shape[0]):
            self.no.push_back(sxz[double](nodes[n, 0],
                                          nodes[n, 1]))
        for n in range(triangles.shape[0]):
            self.tri.push_back(triangleElem[uint32_t](triangles[n, 0],
                                                      triangles[n, 1],
                                                      triangles[n, 2]))

        if method == 'FSM':
            xmin = np.min(nodes[:, 0])
            xmax = np.max(nodes[:, 0])
            zmin = np.min(nodes[:, 1])
            zmax = np.max(nodes[:, 1])
            pts_ref.push_back(sxz[double](xmin, zmin))
            pts_ref.push_back(sxz[double](xmin, zmax))
            pts_ref.push_back(sxz[double](xmax, zmin))
            pts_ref.push_back(sxz[double](xmax, zmax))

        if cell_slowness:
            if method == 'FSM':
                self.method = b'f'
                self.grid = new Grid2Ducfs[double,uint32_t,Node2Dc[double,uint32_t],sxz[double]](self.no,
                                                                                                 self.tri,
                                                                                                 eps,
                                                                                                 maxit,
                                                                                                 pts_ref, 2,
                                                                                                 n_threads,
                                                                                                 process_obtuse)
            elif method == 'SPM':
                self.method = b's'
                self.grid = new Grid2Ducsp[double,uint32_t,Node2Dcsp[double,uint32_t],sxz[double]](self.no,
                                                                                                   self.tri,
                                                                                                   n_secondary,
                                                                                                   n_threads)
            else:
                raise ValueError('Method {0:s} undefined'.format(method))
        else:
            if method == 'FSM':
                self.method = b'f'
                self.grid = new Grid2Dunfs[double,uint32_t,Node2Dn[double,uint32_t],sxz[double]](self.no,
                                                                                                 self.tri,
                                                                                                 eps,
                                                                                                 maxit,
                                                                                                 pts_ref, 2,
                                                                                                 n_threads,
                                                                                                 process_obtuse)
            elif method == 'SPM':
                self.method = b's'
                self.grid = new Grid2Dunsp[double,uint32_t,Node2Dnsp[double,uint32_t],sxz[double]](self.no,
                                                                                                   self.tri,
                                                                                                   n_secondary,
                                                                                                   n_threads)
            else:
                raise ValueError('Method {0:s} undefined'.format(method))

    def __dealloc__(self):
        del self.grid

    def __reduce__(self):
        if self.method == b'f':
            method = 'FSM'
        elif self.method == b's':
            method = 'SPM'

        nodes = np.ndarray((self.no.size(), 2))
        triangles = np.ndarray((self.tri.size(), 3), dtype=int)
        cdef int n
        cdef int nn
        for n in range(nodes.shape[0]):
            nodes[n, 0] = self.no[n].x
            nodes[n, 1] = self.no[n].z
        for n in range(triangles.shape[0]):
            for nn in range(3):
                triangles[n, nn] = self.tri[n].i[nn]

        constructor_params = (nodes, triangles,
                              method, self.cell_slowness,
                              self._n_threads,
                              self.eps, self.maxit, self.process_obtuse,
                              self.n_secondary)
        return (_rebuild2d, constructor_params)

    @property
    def n_threads(self):
        """int: number of threads for raytracing"""
        return self._n_threads

    @property
    def nparams(self):
        """int: total number of parameters for mesh"""
        if self.cell_slowness:
            return self.tri.size()
        else:
            return self.no.size()

    def get_number_of_nodes(self):
        """
        Returns
        -------
        int:
            number of nodes in grid
        """
        return self.no.size()

    def get_number_of_cells(self):
        """
        Returns
        -------
        int:
            number of cells in grid
        """
        return self.tri.size()

    def get_grid_traveltimes(self, thread_no=0):
        """
        get_grid_traveltimes(thread_no=0)

        Obtain traveltimes computed at primary grid nodes

        Parameters
        ----------
        thread_no : int
            thread used to computed traveltimes (default is 0)

        Returns
        -------
        tt: np ndarray, shape (nnodes,)
            traveltimes
        """
        if thread_no >= self._n_threads:
            raise ValueError('Thread number is larger than number of threads')
        cdef vector[double] tmp
        cdef int n
        self.grid.getTT(tmp, thread_no)
        tt = np.empty((tmp.size(),))
        for n in range(tmp.size()):
            tt[n] = tmp[n]
        return tt

    def set_slowness(self, slowness):
        """
        set_slowness(slowness)

        Assign slowness to grid

        Parameters
        ----------
        slowness : np ndarray, shape (nparams, )
        """
        if slowness.size != self.nparams:
            raise ValueError('Slowness vector has wrong size')

        if not slowness.flags['C_CONTIGUOUS']:
            slowness = np.ascontiguousarray(slowness)

        cdef vector[double] slown
        cdef int i
        for i in range(slowness.size):
            slown.push_back(slowness[i])
        self.grid.setSlowness(slown)

    def set_velocity(self, velocity):
        """
        set_velocity(velocity)

        Assign velocity to grid

        Parameters
        ----------
        velocity : np ndarray, shape (nparams, )
        """
        if velocity.size != self.nparams:
            raise ValueError('velocity vector has wrong size')

        if not velocity.flags['C_CONTIGUOUS']:
            velocity = np.ascontiguousarray(velocity)

        cdef vector[double] slown
        cdef int i
        for i in range(velocity.size):
            slown.push_back(1./velocity[i])
        self.grid.setSlowness(slown)

    def raytrace(self, source, rcv, slowness=None, thread_no=None,
                 aggregate_src=False, return_rays=False):
        """
        raytrace(source, rcv, slowness=None, thread_no=None, aggregate_src=False, return_rays=False) -> tt, rays

        Perform raytracing

        Parameters
        ----------
        source : 2D np.ndarray with 2 or 3 columns
            see notes below
        rcv : 2D np.ndarray with 2 columns
            Columns correspond to x and z coordinates
        slowness : np ndarray, (None by default)
            slowness at grid nodes or cells (depending on cell_slowness)
            if None, slowness must have been assigned previously
        thread_no : int (None by default)
            Perform calculations in thread number "thread_no"
            if None, attempt to run in parallel if warranted by number of
            sources and value of n_threads in constructor
        aggregate_src : bool (False by default)
            if True, all source coordinates belong to a single event
        return_rays : bool (False by default)
            Return raypaths

        Returns
        -------
        tt : np.ndarray
            travel times for the appropriate source-rcv  (see Notes below)
        rays : :obj:`list` of :obj:`np.ndarray`
            Coordinates of segments forming raypaths (if return_rays is True)

        Notes
        -----
        If source has 2 columns:
            - Columns correspond to x and z coordinates
            - Origin time (t0) is 0 for all points
        If source has 3 columns:
            - 1st column corresponds to origin times
            - 2nd & 3rd columns correspond to x and z coordinates

        source and rcv can contain the same number of rows, each row
        corresponding to a source-receiver pair, or the number of rows may
        differ if aggregate_src is True or if all rows in source are identical.
        """

        # check input data consistency

        if source.ndim != 2 or rcv.ndim != 2:
            raise ValueError('source and rcv should be 2D arrays')

        if source.shape[1] == 2:
            src = source
            Tx = np.unique(source, axis=0)
            t0 = np.zeros((Tx.shape[0], 1))
            nTx = Tx.shape[0]
        elif source.shape[1] == 3:
            src = source[:,1:3]
            tmp = np.unique(source, axis=0)
            nTx = tmp.shape[0]
            Tx = tmp[:,1:3]
            t0 = tmp[:,0]
        else:
            raise ValueError('source should be either nsrc x 2 or 3')

        if src.shape[1] != 2 or rcv.shape[1] != 2:
            raise ValueError('src and rcv should be ndata x 2')

        if slowness is not None:
            self.set_slowness(slowness)

        cdef vector[vector[sxz[double]]] vTx
        cdef vector[vector[sxz[double]]] vRx
        cdef vector[vector[double]] vt0
        cdef vector[vector[double]] vtt

        cdef vector[vector[vector[sxz[double]]]] r_data
        cdef size_t thread_nb

        cdef int i, n, n2, nt

        vTx.resize(nTx)
        vRx.resize(nTx)
        vt0.resize(nTx)
        vtt.resize(nTx)
        if return_rays:
            r_data.resize(nTx)

        iRx = []
        if nTx == 1:
            vTx[0].push_back(sxz[double](src[0,0], src[0,1]))
            for r in rcv:
                vRx[0].push_back(sxz[double](r[0], r[1]))
            vt0[0].push_back(t0[0])
            vtt[0].resize(rcv.shape[0])
            iRx.append(np.arange(rcv.shape[0]))
        elif aggregate_src:
            for t in Tx:
                vTx[0].push_back(sxz[double](t[0], t[1]))
            for t in t0:
                vt0[0].push_back(t)
            for r in rcv:
                vRx[0].push_back(sxz[double](r[0], r[1]))
            vtt[0].resize(rcv.shape[0])
            nTx = 1
            iRx.append(np.arange(rcv.shape[0]))
        else:
            if src.shape != rcv.shape:
                raise ValueError('src and rcv should be of equal size')

            for n in range(nTx):
                ind = np.sum(Tx[n,:] == src, axis=1) == 2
                iRx.append(np.nonzero(ind)[0])
                vTx[n].push_back(sxz[double](Tx[n,0], Tx[n,1]))
                vt0[n].push_back(t0[n])
                for r in rcv[ind,:]:
                    vRx[n].push_back(sxz[double](r[0], r[1]))
                vtt[n].resize(vRx[n].size())

        tt = np.zeros((rcv.shape[0],))
        if nTx < self._n_threads or self._n_threads == 1:
            if return_rays==False:
                for n in range(nTx):
                    self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], 0)
            else:
                for n in range(nTx):
                    self.grid.raytrace(vTx[n], vt0[n], vRx[n], vtt[n], r_data[n], 0)

        elif thread_no is not None:
            # we should be here for just one event
            assert nTx == 1
            thread_nb = thread_no

            if return_rays:
                self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], r_data[0], thread_nb)
                for nt in range(vtt[0].size()):
                    tt[nt] = vtt[0][nt]
                rays = []
                for n2 in range(vRx.size()):
                    r = np.empty((r_data[0][n2].size(), 2))
                    for nn in range(r_data[0][n2].size()):
                        r[nn, 0] = r_data[0][n2][nn].x
                        r[nn, 1] = r_data[0][n2][nn].z
                    rays.append(r)
                return tt, rays

            else:
                self.grid.raytrace(vTx[0], vt0[0], vRx[0], vtt[0], thread_nb)
                for nt in range(vtt[0].size()):
                    tt[nt] = vtt[0][nt]
                return tt

        else:
            if return_rays==False:
                self.grid.raytrace(vTx, vt0, vRx, vtt)
            else:
                self.grid.raytrace(vTx, vt0, vRx, vtt, r_data)

        for n in range(nTx):
            for nt in range(vtt[n].size()):
                tt[iRx[n][nt]] = vtt[n][nt]

        if return_rays:
            rays = [ [0.0] for n in range(rcv.shape[0])]
            for n in range(nTx):
                r = [ [0.0] for i in range(vRx[n].size())]
                for n2 in range(vRx[n].size()):
                    r[n2] = np.empty((r_data[n][n2].size(), 2))
                    for nn in range(r_data[n][n2].size()):
                        r[n2][nn, 0] = r_data[n][n2][nn].x
                        r[n2][nn, 1] = r_data[n][n2][nn].z
                for nt in range(vtt[n].size()):
                    rays[iRx[n][nt]] = r[nt]

        if return_rays==False:
            return tt
        else:
            return tt, rays

    def to_vtk(self, fields, filename):
        """
        to_vtk(fields, filename)

        Save mesh variables and/or raypaths to VTK format

        Parameters
        ----------
        fields: dict
            dict of variables to save to file. Variables should be np.ndarray of
            size equal to either the number of nodes of the number of cells of
            the mesh, or a list of raypath coordinates.
        filename: str
            Name of file without extension for saving (extension vtu will be
            added).  Raypaths are saved in separate files, and filename will
            be appended by the dict key and have a vtp extension.

        Notes
        -----
        VTK files can be visualized with Paraview (https://www.paraview.org)
        """
        cdef int n, nn
        ugrid = vtk.vtkUnstructuredGrid()
        tPts = vtk.vtkPoints()
        tPts.SetNumberOfPoints(self.no.size())
        for n in range(self.no.size()):
            tPts.InsertPoint(n, self.no[n].x, 0.0, self.no[n].z)
        ugrid.SetPoints(tPts)
        tri = vtk.vtkTriangle
        for n in range(self.tri.size()):
            for nn in range(3):
                tri.GetPointIds().SetId(nn, self.tet[n].i[nn])
            ugrid.InsertNextCell(tri.GetCellType(), tri.GetPointIds())

        save_grid = False
        for fn in fields:
            data = fields[fn]

            if isinstance(data, list):
                self._save_raypaths(data, filename+'_'+fn+'.vtp')
            else:
                save_grid = True
                scalar = vtk.vtkDoubleArray()
                scalar.SetName(fn)
                scalar.SetNumberOfComponents(1)
                scalar.SetNumberOfTuples(data.size)
                if data.size == self.get_number_of_nodes():
                    for n in range(data.size):
                        scalar.SetTuple1(n, data[n])
                    ugrid.GetPointData().AddArray(scalar)
                elif data.size == self.get_number_of_cells():
                    for n in range(data.size):
                        scalar.SetTuple1(n, data[n])
                    ugrid.GetCellData().AddArray(scalar)
                else:
                    raise ValueError('Field {0:s} has incorrect size'.format(fn))

        if save_grid:
            writer = vtk.vtkXMLUnstructuredGridWriter()
            writer.SetFileName(filename+'.vtu')
            writer.SetInputData(ugrid)
            writer.SetDataModeToBinary()
            writer.Update()

    def  _save_raypaths(self, rays, filename):
        polydata = vtk.vtkPolyData()
        cellarray = vtk.vtkCellArray()
        pts = vtk.vtkPoints()
        npts = 0
        for n in range(len(rays)):
            npts += rays[n].shape[0]
        pts.SetNumberOfPoints(npts)
        npts = 0
        for n in range(len(rays)):
            for p in range(rays[n].shape[0]):
                pts.InsertPoint(npts, rays[n][p, 0], 0.0, rays[n][p, 2])
                npts += 1
        polydata.SetPoints(pts)
        npts = 0
        for n in range(len(rays)):
            line = vtk.vtkPolyLine()
            line.GetPointIds().SetNumberOfIds(rays[n].shape[0])
            for p in range(rays[n].shape[0]):
                line.GetPointIds().SetId(p, npts)
                npts += 1
            cellarray.InsertNextCell(line)
        polydata.SetLines(cellarray)
        writer = vtk.vtkXMLPolyDataWriter()
        writer.SetFileName(filename)
        writer.SetInputData(polydata)
        writer.SetDataModeToBinary()
        writer.Update()

    @staticmethod
    def builder(filename, size_t n_threads=1, bool cell_slowness=1,
                str method='FSM',double eps=1.e-15, int maxit=20,
                bool process_obtuse=1,uint32_t n_secondary=5):
        """
        builder(filename, n_threads, cell_slowness, method, eps, maxit, process_obtuse, n_secondary)

        Build instance of Mesh2d from VTK file

        Parameters
        ----------
        filename : str
            Name of file holding a vtkUnstructuredGrid.
            The grid must have point or cell attribute named either
            'Slowness', 'slowness', 'Velocity', 'velocity', or
            'P-wave velocity'.  All cells must be of type vtkTriangle

        Other parameters are defined in Constructor

        Returns
        -------
        mesh: :obj:`Mesh2d`
            mesh instance
        """

        cdef int n, nn
        reader = vtk.vtkXMLUnstructuredGridReader()
        reader.SetFileName(filename)
        reader.Update()

        data = reader.GetOutput()
        nodes = numpy_support.vtk_to_numpy(data.GetPoints().GetData())
        nod = np.ascontiguousarray(nodes[:,[0, 2]])
        for n in range(data.GetNumberOfCells()):
            if data.GetCellType(n) != vtk.VTK_TRIANGLE:
                raise ValueError('{0:s} should only contain triangles')
        tri = np.ndarray((data.GetNumberOfCells(), 3))
        for n in range(data.GetNumberOfCells()):
            for nn in range(3):
                tri[n, nn] = data.GetCell(n).GetPointIds().GetId(nn)

        names = ('Slowness', 'slowness', 'Velocity', 'velocity',
                 'P-wave velocity')
        for name in names:
            if data.GetPointData().HasArray(name):
                cell_slowness = 0
                data = numpy_support.vtk_to_numpy(data.GetPointData().GetArray(name))
                break
            if data.GetCellData().HasArray(name):
                cell_slowness = 1
                data = numpy_support.vtk_to_numpy(data.GetCellData().GetArray(name))
                break
        else:
            raise ValueError('File should contain slowness or velocity data')

        if 'lowness' in name:
            slowness = data
        else:
            slowness = 1.0 / data

        m = Mesh2d(nod, tri, n_threads, cell_slowness, method, eps, maxit,
                   process_obtuse, n_secondary)
        m.set_slowness(slowness)
        return m


def _rebuild3d(constructor_params):
    (nodes, tetra, method, cell_slowness, n_threads, tt_from_rp, interp_vel, eps,
     maxit, gradient_method, min_dist, n_secondary, n_tertiary,
     radius_tertiary) = constructor_params

    g = Mesh3d(nodes, tetra, n_threads, cell_slowness, method, gradient_method,
               tt_from_rp, interp_vel, eps, maxit, min_dist, n_secondary,
               n_tertiary, radius_tertiary)
    return g

def _rebuild2d(constructor_params):
    (nodes, triangles, method, cell_slowness, n_threads, eps, maxit,
     process_obtuse, n_secondary) = constructor_params

    g = Mesh2d(nodes, triangles, n_threads, cell_slowness, method, eps, maxit,
        process_obtuse, n_secondary)
    return g
